import dataclasses
import datetime
import enum
import functools
import pytest

from pywa import handlers, types, WhatsApp
from pywa_async import WhatsApp as WhatsAppAsync
from pywa.handlers import Handler

FAKE_WA = WhatsApp(phone_id="1234567890", token="1234567890:1234567890")
FAKE_WA_ASYNC = WhatsAppAsync(phone_id="1234567890", token="1234567890:1234567890")

CallbackButtonOnlyDataIsNeeded = functools.partial(
    types.CallbackButton,
    _client=FAKE_WA,
    raw={},
    id="123",
    type=types.MessageType.INTERACTIVE,
    metadata=types.Metadata(
        display_phone_number="1234567890",
        phone_number_id="1234567890",
    ),
    from_user=types.User(
        wa_id="1234567890",
        name="John",
    ),
    timestamp=datetime.datetime.now(),
    reply_to_message=types.ReplyToMessage(
        message_id="123",
        from_user_id="1234567890",
        referred_product=None,
    ),
    title="Click me",
)


def test_all_updates_are_overridden():
    assert (
        FAKE_WA._handlers_to_update_constractor.keys()
        == FAKE_WA_ASYNC._handlers_to_update_constractor.keys()
    )


def test_safe_issubclas():
    assert not handlers._safe_issubclass("None", str)


@pytest.mark.asyncio
async def test_resolve_factory_iterable_with_callback_data_subclasses():
    @dataclasses.dataclass
    class User(types.CallbackData):
        id: int
        name: str

    @dataclasses.dataclass
    class Post(types.CallbackData):
        id: int
        title: str

    tuple_factory = (User, Post)
    callback_data = (User(id=1, name="John"), Post(id=1, title="Hello"))
    constructor, factory_filter = handlers._resolve_factory(tuple_factory, "data")
    button_data = types.Button(title="Click me", callback_data=callback_data).to_dict()[
        "reply"
    ]["id"]
    constructed_data = await constructor(button_data)
    assert constructed_data == callback_data

    update = CallbackButtonOnlyDataIsNeeded(data=button_data)
    assert factory_filter(FAKE_WA, update)


@pytest.mark.asyncio
async def test_resolve_factory_iterable_with_just_one_callback_data_subclass():
    @dataclasses.dataclass
    class User(types.CallbackData):
        id: int
        name: str

    @dataclasses.dataclass
    class User2:
        id: int
        name: str

        @classmethod
        async def from_split(cls, data: str):
            _id, name = data.split(":")
            return cls(int(_id), name)

        def to_str(self):
            return f"{self.id}:{self.name}"

    callback_data_constructed = [User(id=1, name="John"), User2(1, "Hello")]
    constructor, factory_filter = handlers._resolve_factory(
        [User, User2.from_split], "data"
    )
    button_data = types.Button(
        title="Click me",
        callback_data=[
            callback_data_constructed[0],
            callback_data_constructed[1].to_str(),
        ],
    ).to_dict()["reply"]["id"]
    constructed_data = await constructor(button_data)
    assert constructed_data == callback_data_constructed

    update = CallbackButtonOnlyDataIsNeeded(data=button_data)
    assert factory_filter(FAKE_WA, update)


@pytest.mark.asyncio
async def test_resolve_factory_iterable_without_callback_data_subclasses():
    @dataclasses.dataclass
    class User:
        id: int
        name: str

        @classmethod
        async def from_split(cls, data: str):
            _id, name = data.split(":")
            return cls(int(_id), name)

        def to_str(self):
            return f"{self.id}:{self.name}"

    @dataclasses.dataclass
    class Post:
        id: int
        title: str

        @classmethod
        async def from_split(cls, data: str):
            _id, title = data.split(":")
            return cls(int(_id), title)

        def to_str(self):
            return f"{self.id}:{self.title}"

    callback_data_constructed = [User(1, "John"), Post(1, "Hello")]
    constructor, factory_filter = handlers._resolve_factory(
        [User.from_split, Post.from_split], "data"
    )
    assert factory_filter is None
    button_data = types.Button(
        title="Click me",
        callback_data=f"{callback_data_constructed[0].to_str()}"
        f"{types.CallbackData.__callback_sep__}"
        f"{callback_data_constructed[1].to_str()}",
    ).to_dict()["reply"]["id"]
    constructed_data = await constructor(button_data)
    assert constructed_data == callback_data_constructed


def test_resolve_factory_callback_data_subclass():
    @dataclasses.dataclass
    class User(types.CallbackData):
        id: int
        name: str

    callback_data = User(id=1, name="John")
    constructor, factory_filter = handlers._resolve_factory(User, "data")
    button_data = types.Button(title="Click me", callback_data=callback_data).to_dict()[
        "reply"
    ]["id"]
    constructed_data = constructor(button_data)
    assert constructed_data == callback_data

    update = CallbackButtonOnlyDataIsNeeded(data=button_data)
    assert factory_filter(FAKE_WA, update)


def test_resolve_factory_custom():
    class UserType(str, enum.Enum):
        USER = enum.auto()
        ADMIN = enum.auto()

    constructor, factory_filter = handlers._resolve_factory(UserType, "data")
    assert factory_filter is None
    button_data = types.Button(title="Click me", callback_data=UserType.USER).to_dict()[
        "reply"
    ]["id"]
    constructed_data = constructor(button_data)
    assert constructed_data == UserType.USER

    def to_int(data: str):
        return int(data)

    constructor, factory_filter = handlers._resolve_factory(to_int, "data")
    assert factory_filter is None
    button_data = types.Button(title="Click me", callback_data="123").to_dict()[
        "reply"
    ]["id"]
    constructed_data = constructor(button_data)
    assert constructed_data == 123


@pytest.mark.asyncio
async def test_get_factored_update_factory_before_filters_false():
    @dataclasses.dataclass
    class User(types.CallbackData):
        id: int
        name: str

    pass_filter = lambda _, b: b.data.startswith(str(User.__callback_id__))
    not_pass_filter = lambda _, b: not b.data.startswith(str(User.__callback_id__))

    handler = handlers.CallbackButtonHandler(
        lambda _, __: None,
        pass_filter,
        factory=User,
    )
    assert handler.filters
    assert handler.factory_filter is not None
    assert handler.factory is not None

    update = CallbackButtonOnlyDataIsNeeded(
        data=types.Button(
            title="Click me", callback_data=User(id=1, name="John")
        ).to_dict()["reply"]["id"],
    )

    assert handler.factory_filter(None, update)
    assert handler.filters[0](FAKE_WA, update)

    constructed_update = await handlers._get_factored_update(
        handler=handler,
        wa=FAKE_WA,
        update=update,
        field_name="data",
    )
    assert constructed_update.data == User(id=1, name="John")

    handler.filters = (not_pass_filter,)
    assert not handler.filters[0](None, update)
    assert handler.factory_filter(None, update)
    should_be_none = await handlers._get_factored_update(
        handler=handler,
        wa=FAKE_WA,
        update=update,
        field_name="data",
    )
    assert should_be_none is None


@pytest.mark.asyncio
async def test_get_factored_update_factory_before_filters_true():
    @dataclasses.dataclass
    class User(types.CallbackData):
        id: int
        name: str

    pass_filter = lambda _, b: b.data.id == 1
    not_pass_filter = lambda _, b: b.data.id != 1

    handler = handlers.CallbackButtonHandler(
        lambda _, __: None,
        pass_filter,
        factory=User,
        factory_before_filters=True,
    )
    assert handler.filters
    assert handler.factory_filter is not None
    assert handler.factory is not None

    update = CallbackButtonOnlyDataIsNeeded(
        data=types.Button(
            title="Click me", callback_data=User(id=1, name="John")
        ).to_dict()["reply"]["id"],
    )

    assert handler.factory_filter(None, update)

    constructed_update = await handlers._get_factored_update(
        handler=handler,
        wa=FAKE_WA,
        update=update,
        field_name="data",
    )
    assert constructed_update.data == User(id=1, name="John")
    assert handler.factory_filter(None, update)

    handler.filters = (not_pass_filter,)
    should_be_none = await handlers._get_factored_update(
        handler=handler,
        wa=FAKE_WA,
        update=update,
        field_name="data",
    )
    assert should_be_none is None


@pytest.mark.asyncio
async def test_access_to_attr_from_filter_when_factory_before_filters_false():
    @dataclasses.dataclass
    class User(types.CallbackData):
        id: int
        name: str

    handler = handlers.CallbackButtonHandler(
        lambda _, __: None,
        lambda _, b: b.data.id == 1,
        factory=User,
        factory_before_filters=False,
    )
    update = CallbackButtonOnlyDataIsNeeded(
        data=types.Button(
            title="Click me", callback_data=User(id=1, name="John")
        ).to_dict()["reply"]["id"],
    )

    with pytest.raises(AttributeError):
        await handlers._get_factored_update(
            handler=handler,
            wa=FAKE_WA,
            update=update,
            field_name="data",
        )



import json

import pytest

from pywa.types.flows import (
    FlowJSON,
    Screen,
    Layout,
    LayoutType,
    Form,
    TextInput,
    InputType,
    Footer,
    Action,
    FlowActionType,
    ActionNext,
    ActionNextType,
    TextHeading,
    TextSubheading,
    RadioButtonsGroup,
    DataSource,
    TextArea,
    Dropdown,
    CheckboxGroup,
    TextBody,
    OptIn,
    EmbeddedLink,
    DataKey,
    FormRef,
    ScreenData,
    FlowResponse,
)
from pywa.utils import Version

FLOWS_VERSION = "2.1"

customer_satisfaction_survey = FlowJSON(
    version="2.1",
    screens=[
        Screen(
            id="RECOMMEND",
            title="Feedback 1 of 2",
            data={},
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="form",
                        children=[
                            TextSubheading(text="Would you recommend us to a friend?"),
                            recommend_radio := RadioButtonsGroup(
                                name="recommend_radio",
                                label="Choose one",
                                data_source=[
                                    DataSource(id="0", title="Yes"),
                                    DataSource(id="1", title="No"),
                                ],
                                required=True,
                            ),
                            TextSubheading(text="How could we do better?"),
                            comment_text := TextArea(
                                name="comment_text",
                                label="Leave a comment",
                                required=False,
                            ),
                            Footer(
                                label="Continue",
                                on_click_action=Action(
                                    name=FlowActionType.NAVIGATE,
                                    next=ActionNext(
                                        type=ActionNextType.SCREEN, name="RATE"
                                    ),
                                    payload={
                                        "recommend_radio": recommend_radio.form_ref,
                                        "comment_text": comment_text.form_ref,
                                    },
                                ),
                            ),
                        ],
                    )
                ],
            ),
        ),
        Screen(
            id="RATE",
            title="Feedback 2 of 2",
            data=[
                recommend_radio := ScreenData(key="recommend_radio", example="Example"),
                comment_text := ScreenData(key="comment_text", example="Example"),
            ],
            terminal=True,
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="form",
                        children=[
                            TextSubheading(text="Rate the following: "),
                            purchase_rating := Dropdown(
                                name="purchase_rating",
                                label="Purchase experience",
                                required=True,
                                data_source=[
                                    DataSource(id="0", title="★★★★★ • Excellent (5/5)"),
                                    DataSource(id="1", title="★★★★☆ • Good (4/5)"),
                                    DataSource(id="2", title="★★★☆☆ • Average (3/5)"),
                                    DataSource(id="3", title="★★☆☆☆ • Poor (2/5)"),
                                    DataSource(id="4", title="★☆☆☆☆ • Very Poor (1/5)"),
                                ],
                            ),
                            delivery_rating := Dropdown(
                                name="delivery_rating",
                                label="Delivery and setup",
                                required=True,
                                data_source=[
                                    DataSource(id="0", title="★★★★★ • Excellent (5/5)"),
                                    DataSource(id="1", title="★★★★☆ • Good (4/5)"),
                                    DataSource(id="2", title="★★★☆☆ • Average (3/5)"),
                                    DataSource(id="3", title="★★☆☆☆ • Poor (2/5)"),
                                    DataSource(id="4", title="★☆☆☆☆ • Very Poor (1/5)"),
                                ],
                            ),
                            cs_rating := Dropdown(
                                name="cs_rating",
                                label="Customer service",
                                required=True,
                                data_source=[
                                    DataSource(id="0", title="★★★★★ • Excellent (5/5)"),
                                    DataSource(id="1", title="★★★★☆ • Good (4/5)"),
                                    DataSource(id="2", title="★★★☆☆ • Average (3/5)"),
                                    DataSource(id="3", title="★★☆☆☆ • Poor (2/5)"),
                                    DataSource(id="4", title="★☆☆☆☆ • Very Poor (1/5)"),
                                ],
                            ),
                            Footer(
                                label="Done",
                                on_click_action=Action(
                                    name=FlowActionType.COMPLETE,
                                    payload={
                                        "purchase_rating": purchase_rating.form_ref,
                                        "delivery_rating": delivery_rating.form_ref,
                                        "cs_rating": cs_rating.form_ref,
                                        "recommend_radio": recommend_radio.data_key,
                                        "comment_text": comment_text.data_key,
                                    },
                                ),
                            ),
                        ],
                    )
                ],
            ),
        ),
    ],
)

load_re_engagement = FlowJSON(
    version="2.1",
    screens=[
        Screen(
            id="SIGN_UP",
            title="Finish Sign Up",
            terminal=True,
            data={},
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="form",
                        children=[
                            first_name := TextInput(
                                name="firstName",
                                label="First Name",
                                input_type=InputType.TEXT,
                                required=True,
                                visible=True,
                            ),
                            last_name := TextInput(
                                name="lastName",
                                label="Last Name",
                                input_type=InputType.TEXT,
                                required=True,
                                visible=True,
                            ),
                            email := TextInput(
                                name="email",
                                label="Email Address",
                                input_type=InputType.EMAIL,
                                required=True,
                                visible=True,
                            ),
                            Footer(
                                label="Done",
                                enabled=True,
                                on_click_action=Action(
                                    name=FlowActionType.COMPLETE,
                                    payload={
                                        "firstName": first_name.form_ref,
                                        "lastName": last_name.form_ref,
                                        "email": email.form_ref,
                                    },
                                ),
                            ),
                        ],
                    )
                ],
            ),
        )
    ],
)

costumer_engagement = FlowJSON(
    version="2.1",
    screens=[
        Screen(
            id="QUESTION_ONE",
            title="Question 1 of 3",
            data={},
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="form",
                        children=[
                            TextHeading(
                                text="You've found the perfect deal, what do you do next?"
                            ),
                            question1_checkbox := CheckboxGroup(
                                name="question1Checkbox",
                                label="Choose all that apply:",
                                required=True,
                                data_source=[
                                    DataSource(id="0", title="Buy it right away"),
                                    DataSource(
                                        id="1", title="Check reviews before buying"
                                    ),
                                    DataSource(
                                        id="2", title="Share it with friends + family"
                                    ),
                                    DataSource(
                                        id="3", title="Buy multiple, while its cheap"
                                    ),
                                    DataSource(id="4", title="None of the above"),
                                ],
                            ),
                            Footer(
                                label="Continue",
                                on_click_action=Action(
                                    name=FlowActionType.NAVIGATE,
                                    next=ActionNext(
                                        type=ActionNextType.SCREEN, name="QUESTION_TWO"
                                    ),
                                    payload={
                                        "question1Checkbox": question1_checkbox.form_ref
                                    },
                                ),
                            ),
                        ],
                    )
                ],
            ),
        ),
        Screen(
            id="QUESTION_TWO",
            title="Question 2 of 3",
            data=[
                question1_checkbox := ScreenData(
                    key="question1Checkbox", example=["Example", "Example2"]
                ),
            ],
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="form",
                        children=[
                            TextHeading(
                                text="Its your birthday in two weeks, how might you prepare?"
                            ),
                            question2_radio_buttons := RadioButtonsGroup(
                                name="question2RadioButtons",
                                label="Choose all that apply:",
                                required=True,
                                data_source=[
                                    DataSource(id="0", title="Buy something new"),
                                    DataSource(id="1", title="Wear the same, as usual"),
                                    DataSource(id="2", title="Look for a deal online"),
                                ],
                            ),
                            Footer(
                                label="Continue",
                                on_click_action=Action(
                                    name=FlowActionType.NAVIGATE,
                                    next=ActionNext(
                                        type=ActionNextType.SCREEN,
                                        name="QUESTION_THREE",
                                    ),
                                    payload={
                                        "question2RadioButtons": question2_radio_buttons.form_ref,
                                        "question1Checkbox": question1_checkbox.data_key,
                                    },
                                ),
                            ),
                        ],
                    )
                ],
            ),
        ),
        Screen(
            id="QUESTION_THREE",
            title="Question 3 of 3",
            data=[
                question2_radio_buttons := ScreenData(
                    key="question2RadioButtons", example="Example"
                ),
                question1_checkbox := ScreenData(
                    key="question1Checkbox", example=["Example", "Example2"]
                ),
            ],
            terminal=True,
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="form",
                        children=[
                            TextHeading(text="What's the best gift for a friend?"),
                            question3_checkbox := CheckboxGroup(
                                name="question3Checkbox",
                                label="Choose all that apply:",
                                required=True,
                                data_source=[
                                    DataSource(id="0", title="A gift voucher"),
                                    DataSource(id="1", title="A new outfit "),
                                    DataSource(id="2", title="A bouquet of flowers"),
                                    DataSource(id="3", title="A meal out together"),
                                ],
                            ),
                            Footer(
                                label="Done",
                                on_click_action=Action(
                                    name=FlowActionType.COMPLETE,
                                    payload={
                                        "question1Checkbox": question1_checkbox.data_key,
                                        "question2RadioButtons": question2_radio_buttons.data_key,
                                        "question3Checkbox": question3_checkbox.form_ref,
                                    },
                                ),
                            ),
                        ],
                    )
                ],
            ),
        ),
    ],
)

support_request = FlowJSON(
    version="2.1",
    screens=[
        Screen(
            id="DETAILS",
            title="Get help",
            data={},
            terminal=True,
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="form",
                        children=[
                            name := TextInput(
                                name="name",
                                label="Name",
                                input_type=InputType.TEXT,
                                required=True,
                            ),
                            order_number := TextInput(
                                label="Order number",
                                name="orderNumber",
                                input_type=InputType.NUMBER,
                                required=True,
                                helper_text="",
                            ),
                            topic_radio := RadioButtonsGroup(
                                label="Choose a topic",
                                name="topicRadio",
                                data_source=[
                                    DataSource(id="0", title="Orders and payments"),
                                    DataSource(id="1", title="Maintenance"),
                                    DataSource(id="2", title="Delivery"),
                                    DataSource(id="3", title="Returns"),
                                    DataSource(id="4", title="Other"),
                                ],
                                required=True,
                            ),
                            desc := TextArea(
                                label="Description of issue",
                                required=False,
                                name="description",
                            ),
                            Footer(
                                label="Done",
                                on_click_action=Action(
                                    name=FlowActionType.COMPLETE,
                                    payload={
                                        "name": name.form_ref,
                                        "orderNumber": order_number.form_ref,
                                        "topicRadio": topic_radio.form_ref,
                                        "description": desc.form_ref,
                                    },
                                ),
                            ),
                        ],
                    )
                ],
            ),
        )
    ],
)

communication_preferences = FlowJSON(
    version="2.1",
    screens=[
        Screen(
            id="PREFERENCES",
            title="Update Preferences",
            data={},
            terminal=True,
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="form",
                        children=[
                            communication_types := CheckboxGroup(
                                label="Communication types",
                                required=True,
                                name="communicationTypes",
                                data_source=[
                                    DataSource(
                                        id="0", title="Special offers and promotions"
                                    ),
                                    DataSource(
                                        id="1", title="Changes to my subscription"
                                    ),
                                    DataSource(id="2", title="News and events"),
                                    DataSource(id="3", title="New products"),
                                ],
                            ),
                            contact_prefs := CheckboxGroup(
                                label="Contact Preferences",
                                required=False,
                                name="contactPrefs",
                                data_source=[
                                    DataSource(id="0", title="Whatsapp"),
                                    DataSource(id="1", title="Email"),
                                    DataSource(id="2", title="SMS"),
                                ],
                            ),
                            Footer(
                                label="Done",
                                on_click_action=Action(
                                    name=FlowActionType.COMPLETE,
                                    payload={
                                        "communicationTypes": communication_types.form_ref,
                                        "contactPrefs": contact_prefs.form_ref,
                                    },
                                ),
                            ),
                        ],
                    )
                ],
            ),
        )
    ],
)

register_for_an_event = FlowJSON(
    version="2.1",
    screens=[
        Screen(
            id="SIGN_UP",
            title="Sign Up",
            data={},
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="form",
                        children=[
                            TextHeading(text="Join our next webinar!"),
                            TextBody(text="First, we'll need a few details from you."),
                            first_name := TextInput(
                                name="firstName",
                                label="First Name",
                                input_type=InputType.TEXT,
                                required=True,
                            ),
                            last_name := TextInput(
                                label="Last Name",
                                name="lastName",
                                input_type=InputType.TEXT,
                                required=True,
                            ),
                            email := TextInput(
                                label="Email Address",
                                name="email",
                                input_type=InputType.EMAIL,
                                required=True,
                            ),
                            Footer(
                                label="Continue",
                                on_click_action=Action(
                                    name=FlowActionType.NAVIGATE,
                                    next=ActionNext(
                                        type=ActionNextType.SCREEN, name="SURVEY"
                                    ),
                                    payload={
                                        "firstName": first_name.form_ref,
                                        "lastName": last_name.form_ref,
                                        "email": email.form_ref,
                                    },
                                ),
                            ),
                        ],
                    )
                ],
            ),
        ),
        Screen(
            id="SURVEY",
            title="Thank you",
            data=[
                first_name := ScreenData(key="firstName", example="Example"),
                last_name := ScreenData(key="lastName", example="Example"),
                email := ScreenData(key="email", example="Example"),
            ],
            terminal=True,
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="form",
                        children=[
                            TextHeading(text="Before you go"),
                            TextBody(text="How did you hear about us?"),
                            source := RadioButtonsGroup(
                                name="source",
                                label="Choose one",
                                required=False,
                                data_source=[
                                    DataSource(id="0", title="Friend's recommendation"),
                                    DataSource(id="1", title="TV advertisement"),
                                    DataSource(id="2", title="Search engine"),
                                    DataSource(id="3", title="Social media"),
                                ],
                            ),
                            Footer(
                                label="Done",
                                on_click_action=Action(
                                    name=FlowActionType.COMPLETE,
                                    payload={
                                        "source": source.form_ref,
                                        "firstName": first_name.data_key,
                                        "lastName": last_name.data_key,
                                        "email": email.data_key,
                                    },
                                ),
                            ),
                        ],
                    )
                ],
            ),
        ),
    ],
)

sign_in = FlowJSON(
    version="2.1",
    data_api_version="3.0",
    data_channel_uri="https://example.com",
    routing_model={
        "SIGN_IN": ["SIGN_UP", "FORGOT_PASSWORD"],
        "SIGN_UP": ["TERMS_AND_CONDITIONS"],
        "FORGOT_PASSWORD": [],
        "TERMS_AND_CONDITIONS": [],
    },
    screens=[
        Screen(
            id="SIGN_IN",
            title="Sign in",
            terminal=True,
            data={},
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="sign_in_form",
                        children=[
                            email := TextInput(
                                name="email",
                                label="Email address",
                                input_type=InputType.EMAIL,
                                required=True,
                            ),
                            password := TextInput(
                                name="password",
                                label="Password",
                                input_type=InputType.PASSWORD,
                                required=True,
                            ),
                            EmbeddedLink(
                                text="Don't have an account? Sign up",
                                on_click_action=Action(
                                    name=FlowActionType.NAVIGATE,
                                    next=ActionNext(
                                        type=ActionNextType.SCREEN, name="SIGN_UP"
                                    ),
                                    payload={},
                                ),
                            ),
                            EmbeddedLink(
                                text="Forgot password",
                                on_click_action=Action(
                                    name=FlowActionType.NAVIGATE,
                                    next=ActionNext(
                                        type=ActionNextType.SCREEN,
                                        name="FORGOT_PASSWORD",
                                    ),
                                    payload={"body": "Example"},
                                ),
                            ),
                            Footer(
                                label="Sign in",
                                on_click_action=Action(
                                    name=FlowActionType.DATA_EXCHANGE,
                                    payload={
                                        "email": email.form_ref,
                                        "password": password.form_ref,
                                    },
                                ),
                            ),
                        ],
                    )
                ],
            ),
        ),
        Screen(
            id="SIGN_UP",
            title="Sign up",
            data={},
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="sign_up_form",
                        children=[
                            first_name := TextInput(
                                name="first_name",
                                label="First Name",
                                input_type=InputType.TEXT,
                                required=True,
                            ),
                            last_name := TextInput(
                                name="last_name",
                                label="Last Name",
                                input_type=InputType.TEXT,
                                required=True,
                            ),
                            email := TextInput(
                                name="email",
                                label="Email address",
                                input_type=InputType.EMAIL,
                                required=True,
                            ),
                            password := TextInput(
                                name="password",
                                label="Set password",
                                input_type=InputType.PASSWORD,
                                required=True,
                            ),
                            confirm_password := TextInput(
                                name="confirm_password",
                                label="Confirm password",
                                input_type=InputType.PASSWORD,
                                required=True,
                            ),
                            terms_agreement := OptIn(
                                name="terms_agreement",
                                label="I agree with the terms.",
                                required=True,
                                on_click_action=Action(
                                    name=FlowActionType.NAVIGATE,
                                    next=ActionNext(
                                        type=ActionNextType.SCREEN,
                                        name="TERMS_AND_CONDITIONS",
                                    ),
                                    payload={},
                                ),
                            ),
                            offers_acceptance := OptIn(
                                name="offers_acceptance",
                                label="I would like to receive news and offers.",
                            ),
                            Footer(
                                label="Continue",
                                on_click_action=Action(
                                    name=FlowActionType.DATA_EXCHANGE,
                                    payload={
                                        "first_name": first_name.form_ref,
                                        "last_name": last_name.form_ref,
                                        "email": email.form_ref,
                                        "password": password.form_ref,
                                        "confirm_password": confirm_password.form_ref,
                                        "terms_agreement": terms_agreement.form_ref,
                                        "offers_acceptance": offers_acceptance.form_ref,
                                    },
                                ),
                            ),
                        ],
                    )
                ],
            ),
        ),
        Screen(
            id="FORGOT_PASSWORD",
            title="Forgot password",
            terminal=True,
            data=[
                body := ScreenData(
                    key="body",
                    example=(
                        "Enter your email address for your account and we'll send a reset link. "
                        "The single-use link will expire after 24 hours."
                    ),
                ),
            ],
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="forgot_password_form",
                        children=[
                            TextBody(text=body.data_key),
                            email := TextInput(
                                name="email",
                                label="Email address",
                                input_type=InputType.EMAIL,
                                required=True,
                            ),
                            Footer(
                                label="Sign in",
                                on_click_action=Action(
                                    name=FlowActionType.DATA_EXCHANGE,
                                    payload={"email": email.form_ref},
                                ),
                            ),
                        ],
                    )
                ],
            ),
        ),
        Screen(
            id="TERMS_AND_CONDITIONS",
            title="Terms and conditions",
            data={},
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    TextHeading(text="Our Terms"),
                    TextSubheading(text="Data usage"),
                    TextBody(
                        text="Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed vitae odio dui. Praesent ut nulla tincidunt, scelerisque augue malesuada, volutpat lorem. Aliquam iaculis ex at diam posuere mollis. Suspendisse eget purus ac tellus interdum pharetra. In quis dolor turpis. Fusce in porttitor enim, vitae efficitur nunc. Fusce dapibus finibus volutpat. Fusce velit mi, ullamcorper ac gravida vitae, blandit quis ex. Fusce ultrices diam et justo blandit, quis consequat nisl euismod. Vestibulum pretium est sem, vitae convallis justo sollicitudin non. Morbi bibendum purus mattis quam condimentum, a scelerisque erat bibendum. Nullam sit amet bibendum lectus."
                    ),
                    TextSubheading(text="Privacy policy"),
                    TextBody(
                        text="Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed vitae odio dui. Praesent ut nulla tincidunt, scelerisque augue malesuada, volutpat lorem. Aliquam iaculis ex at diam posuere mollis. Suspendisse eget purus ac tellus interdum pharetra. In quis dolor turpis. Fusce in porttitor enim, vitae efficitur nunc. Fusce dapibus finibus volutpat. Fusce velit mi, ullamcorper ac gravida vitae, blandit quis ex. Fusce ultrices diam et justo blandit, quis consequat nisl euismod. Vestibulum pretium est sem, vitae convallis justo sollicitudin non. Morbi bibendum purus mattis quam condimentum, a scelerisque erat bibendum. Nullam sit amet bibendum lectus."
                    ),
                ],
            ),
        ),
    ],
)

register = FlowJSON(
    version="2.1",
    data_api_version="3.0",
    data_channel_uri="https://example.com",
    routing_model={"REGISTER": ["TERMS_AND_CONDITIONS"], "TERMS_AND_CONDITIONS": []},
    screens=[
        Screen(
            id="REGISTER",
            title="Register for an account",
            terminal=True,
            data=[
                error_messages := ScreenData(
                    key="error_messages",
                    example={"confirm_password": "Passwords don't match."},
                ),
            ],
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="register_form",
                        error_messages=error_messages.data_key,
                        children=[
                            first_name := TextInput(
                                name="first_name",
                                required=True,
                                label="First name",
                                input_type="text",
                            ),
                            last_name := TextInput(
                                name="last_name",
                                required=True,
                                label="Last name",
                                input_type="text",
                            ),
                            email := TextInput(
                                name="email",
                                required=True,
                                label="Email address",
                                input_type="email",
                            ),
                            password := TextInput(
                                name="password",
                                required=True,
                                label="Set password",
                                input_type="password",
                            ),
                            confirm_password := TextInput(
                                name="confirm_password",
                                required=True,
                                label="Confirm password",
                                input_type="password",
                            ),
                            terms_agreement := OptIn(
                                name="terms_agreement",
                                label="I agree with the terms.",
                                required=True,
                                on_click_action=Action(
                                    name=FlowActionType.NAVIGATE,
                                    next=ActionNext(
                                        type=ActionNextType.SCREEN,
                                        name="TERMS_AND_CONDITIONS",
                                    ),
                                    payload={},
                                ),
                            ),
                            offers_acceptance := OptIn(
                                name="offers_acceptance",
                                label="I would like to receive news and offers.",
                            ),
                            Footer(
                                label="Continue",
                                on_click_action=Action(
                                    name=FlowActionType.DATA_EXCHANGE,
                                    payload={
                                        "first_name": first_name.form_ref,
                                        "last_name": last_name.form_ref,
                                        "email": email.form_ref,
                                        "password": password.form_ref,
                                        "confirm_password": confirm_password.form_ref,
                                        "terms_agreement": terms_agreement.form_ref,
                                        "offers_acceptance": offers_acceptance.form_ref,
                                    },
                                ),
                            ),
                        ],
                    )
                ],
            ),
        ),
        Screen(
            id="TERMS_AND_CONDITIONS",
            title="Terms and conditions",
            data={},
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    TextHeading(text="Our Terms"),
                    TextSubheading(text="Data usage"),
                    TextBody(
                        text="Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed vitae odio dui. Praesent ut nulla tincidunt, scelerisque augue malesuada, volutpat lorem. Aliquam iaculis ex at diam posuere mollis. Suspendisse eget purus ac tellus interdum pharetra. In quis dolor turpis. Fusce in porttitor enim, vitae efficitur nunc. Fusce dapibus finibus volutpat. Fusce velit mi, ullamcorper ac gravida vitae, blandit quis ex. Fusce ultrices diam et justo blandit, quis consequat nisl euismod. Vestibulum pretium est sem, vitae convallis justo sollicitudin non. Morbi bibendum purus mattis quam condimentum, a scelerisque erat bibendum. Nullam sit amet bibendum lectus."
                    ),
                    TextSubheading(text="Privacy policy"),
                    TextBody(
                        text="Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed vitae odio dui. Praesent ut nulla tincidunt, scelerisque augue malesuada, volutpat lorem. Aliquam iaculis ex at diam posuere mollis. Suspendisse eget purus ac tellus interdum pharetra. In quis dolor turpis. Fusce in porttitor enim, vitae efficitur nunc. Fusce dapibus finibus volutpat. Fusce velit mi, ullamcorper ac gravida vitae, blandit quis ex. Fusce ultrices diam et justo blandit, quis consequat nisl euismod. Vestibulum pretium est sem, vitae convallis justo sollicitudin non. Morbi bibendum purus mattis quam condimentum, a scelerisque erat bibendum. Nullam sit amet bibendum lectus."
                    ),
                ],
            ),
        ),
    ],
)


get_a_quote = FlowJSON(
    version="2.1",
    data_api_version="3.0",
    data_channel_uri="https://example.com",
    routing_model={
        "DETAILS": ["COVER"],
        "COVER": ["QUOTE"],
        "QUOTE": ["TERMS_AND_CONDITIONS"],
        "TERMS_AND_CONDITIONS": [],
    },
    screens=[
        Screen(
            id="DETAILS",
            title="Your details",
            data=[
                city := ScreenData(
                    key="city", example=[DataSource(id="1", title="Light City, SO")]
                ),
            ],
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="details_form",
                        children=[
                            name := TextInput(
                                label="Your name",
                                input_type=InputType.TEXT,
                                name="name",
                                required=True,
                            ),
                            address := TextInput(
                                label="Street address",
                                input_type=InputType.TEXT,
                                name="address",
                                required=True,
                            ),
                            city := Dropdown(
                                label="City, State",
                                name="city",
                                data_source=city.data_key,
                                required=True,
                            ),
                            zip_code := TextInput(
                                label="Zip code",
                                input_type=InputType.TEXT,
                                name="zip_code",
                                required=True,
                            ),
                            country_region := TextInput(
                                label="Country/Region",
                                input_type=InputType.TEXT,
                                name="country_region",
                                required=True,
                            ),
                            Footer(
                                label="Continue",
                                on_click_action=Action(
                                    name=FlowActionType.DATA_EXCHANGE,
                                    payload={
                                        "name": name.form_ref,
                                        "address": address.form_ref,
                                        "city": city.form_ref,
                                        "zip_code": zip_code.form_ref,
                                        "country_region": country_region.form_ref,
                                    },
                                ),
                            ),
                        ],
                    )
                ],
            ),
        ),
        Screen(
            id="COVER",
            title="Your cover",
            data=[
                options := ScreenData(
                    key="options",
                    example=[
                        DataSource(
                            id="1",
                            title="Fire and theft",
                            description="Cover your home against incidents of theft or accidental fires",
                        ),
                        DataSource(
                            id="2",
                            title="Natural disaster",
                            description="Protect your home against disasters including earthquakes, floods and storms",
                        ),
                        DataSource(
                            id="3",
                            title="Liability",
                            description="Protect yourself from legal liabilities that occur from accidents on your property",
                        ),
                    ],
                ),
            ],
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="cover_form",
                        children=[
                            options_form := CheckboxGroup(
                                name="options",
                                data_source=options.data_key,
                                label="Options",
                                required=True,
                            ),
                            Footer(
                                label="Continue",
                                on_click_action=Action(
                                    name=FlowActionType.DATA_EXCHANGE,
                                    payload={"options": options_form.form_ref},
                                ),
                            ),
                        ],
                    )
                ],
            ),
        ),
        Screen(
            id="QUOTE",
            title="Your quote",
            terminal=True,
            data=[
                excess := ScreenData(
                    key="excess", example=[DataSource(id="1", title="$250")]
                ),
                total := ScreenData(key="total", example="$47.98 per month"),
            ],
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    Form(
                        name="quote_form",
                        children=[
                            Dropdown(
                                label="Excess",
                                name="excess",
                                data_source=excess.data_key,
                                on_select_action=Action(
                                    name=FlowActionType.DATA_EXCHANGE,
                                    payload={"excess": FormRef("excess")},
                                ),
                                required=True,
                            ),
                            RadioButtonsGroup(
                                name="payment_options",
                                label="Payment options",
                                data_source=[
                                    DataSource(id="1", title="Monthly"),
                                    DataSource(id="2", title="Annually (Save $115)"),
                                ],
                                on_select_action=Action(
                                    name=FlowActionType.DATA_EXCHANGE,
                                    payload={
                                        "payment_options": FormRef("payment_options")
                                    },
                                ),
                                required=True,
                                init_value="1",
                            ),
                            TextHeading(text=total.data_key),
                            privacy_policy := OptIn(
                                name="privacy_policy",
                                label="Accept our Privacy Policy",
                                required=True,
                                on_click_action=Action(
                                    name=FlowActionType.NAVIGATE,
                                    next=ActionNext(
                                        type=ActionNextType.SCREEN,
                                        name="TERMS_AND_CONDITIONS",
                                    ),
                                    payload={},
                                ),
                            ),
                            Footer(
                                label="Choose quote",
                                on_click_action=Action(
                                    name=FlowActionType.DATA_EXCHANGE,
                                    payload={
                                        "privacy_policy": privacy_policy.form_ref,
                                    },
                                ),
                            ),
                        ],
                    )
                ],
            ),
        ),
        Screen(
            id="TERMS_AND_CONDITIONS",
            title="Terms and conditions",
            data={},
            layout=Layout(
                type=LayoutType.SINGLE_COLUMN,
                children=[
                    TextHeading(text="Our Terms"),
                    TextSubheading(text="Data usage"),
                    TextBody(
                        text="Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed vitae odio dui. Praesent ut nulla tincidunt, scelerisque augue malesuada, volutpat lorem. Aliquam iaculis ex at diam posuere mollis. Suspendisse eget purus ac tellus interdum pharetra. In quis dolor turpis. Fusce in porttitor enim, vitae efficitur nunc. Fusce dapibus finibus volutpat. Fusce velit mi, ullamcorper ac gravida vitae, blandit quis ex. Fusce ultrices diam et justo blandit, quis consequat nisl euismod. Vestibulum pretium est sem, vitae convallis justo sollicitudin non. Morbi bibendum purus mattis quam condimentum, a scelerisque erat bibendum. Nullam sit amet bibendum lectus."
                    ),
                    TextSubheading(text="Privacy policy"),
                    TextBody(
                        text="Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed vitae odio dui. Praesent ut nulla tincidunt, scelerisque augue malesuada, volutpat lorem. Aliquam iaculis ex at diam posuere mollis. Suspendisse eget purus ac tellus interdum pharetra. In quis dolor turpis. Fusce in porttitor enim, vitae efficitur nunc. Fusce dapibus finibus volutpat. Fusce velit mi, ullamcorper ac gravida vitae, blandit quis ex. Fusce ultrices diam et justo blandit, quis consequat nisl euismod. Vestibulum pretium est sem, vitae convallis justo sollicitudin non. Morbi bibendum purus mattis quam condimentum, a scelerisque erat bibendum. Nullam sit amet bibendum lectus."
                    ),
                ],
            ),
        ),
    ],
)

FLOWS = {
    "customer_satisfaction_survey": customer_satisfaction_survey,
    "load_re_engagement": load_re_engagement,
    "costumer_engagement": costumer_engagement,
    "support_request": support_request,
    "communication_preferences": communication_preferences,
    "register_for_an_event": register_for_an_event,
    "sign_in": sign_in,
    "register": register,
    "get_a_quote": get_a_quote,
}


def test_flows_to_json():
    with open(f"tests/data/flows/{FLOWS_VERSION}/examples.json", "r") as f:
        examples = json.load(f)
    for flow_name, flow in FLOWS.items():
        try:
            assert flow.to_dict() == examples[flow_name]
        except AssertionError:
            raise AssertionError(
                f"Flow {flow_name} does not match example\nFlow: {flow}\nJSON: {examples[flow_name]}"
            )


def test_min_version():
    with pytest.raises(ValueError):
        FlowJSON(version="1.0", screens=[])


def test_empty_form():
    with pytest.raises(ValueError):
        Form(name="form", children=[])


def test_action():
    with pytest.raises(ValueError):
        Action(name=FlowActionType.NAVIGATE)

    with pytest.raises(ValueError):
        Action(name=FlowActionType.COMPLETE)


def test_form_ref():
    assert FormRef("test") == "${form.test}"
    assert FormRef("test", screen="START") == "${screen.START.form.test}"
    assert TextInput(name="test", label="Test").form_ref == "${form.test}"
    assert (
        TextInput(name="test", label="Test").form_ref_of(screen="START")
        == "${screen.START.form.test}"
    )


def test_data_key():
    assert DataKey("test") == "${data.test}"
    assert DataKey("test", screen="START") == "${screen.START.data.test}"
    assert ScreenData(key="test", example="Example").data_key == "${data.test}"
    assert (
        ScreenData(key="test", example="Example").data_key_of(screen="START")
        == "${screen.START.data.test}"
    )


def test_init_values():
    text_entry = TextInput(name="test", label="Test", init_value="Example")
    form = Form(name="form", children=[text_entry])
    assert form.init_values == {"test": "Example"}

    # check for duplicate init_values (in the form level and in the children level)
    with pytest.raises(ValueError):
        TextInput(
            name="test", label="Test", init_value="Example", input_type=InputType.NUMBER
        )
        Form(name="form", init_values={"test": "Example"}, children=[text_entry])

    # test that if form has init_values referred to a data_key,
    # the init_values does not fill up from the .children init_value's
    form_with_init_values_as_data_key = Screen(
        id="test",
        title="Test",
        data=[
            init_vals := ScreenData(key="init_vals", example={"test": "Example"}),
        ],
        layout=Layout(
            children=[
                Form(name="form", init_values=init_vals.data_key, children=[text_entry])
            ]
        ),
    )
    assert isinstance(
        form_with_init_values_as_data_key.layout.children[0].init_values, str
    )


def test_error_messages():
    text_entry = TextInput(name="test", label="Test", error_message="Example")
    form = Form(name="form", children=[text_entry])
    assert form.error_messages == {"test": "Example"}

    # check for duplicate error_messages (in the form level and in the children level)
    with pytest.raises(ValueError):
        TextInput(name="test", label="Test", error_message="Example")
        Form(name="form", error_messages={"test": "Example"}, children=[text_entry])

    # test that if form has error_messages referred to a data_key,
    # the error_messages does not fill up from the .children error_message's
    form_with_error_messages_as_data_key = Screen(
        id="test",
        title="Test",
        data=[
            error_msgs := ScreenData(key="error_msgs", example={"test": "Example"}),
        ],
        layout=Layout(
            children=[
                Form(
                    name="form",
                    error_messages=error_msgs.data_key,
                    children=[text_entry],
                )
            ]
        ),
    )
    assert isinstance(
        form_with_error_messages_as_data_key.layout.children[0].error_messages, str
    )


def test_screen_data():
    assert Screen(
        id="test",
        title="Test",
        data=[
            ScreenData(key="str", example="Example"),
            ScreenData(key="int", example=1),
            ScreenData(key="float", example=1.0),
            ScreenData(key="bool", example=True),
        ],
        layout=Layout(children=[]),
    ) == Screen(
        id="test",
        title="Test",
        data={
            "str": {"type": "string", "__example__": "Example"},
            "int": {"type": "number", "__example__": 1},
            "float": {"type": "number", "__example__": 1.0},
            "bool": {"type": "boolean", "__example__": True},
        },
        layout=Layout(children=[]),
    )

    # ---

    screen_1 = Screen(
        id="test",
        title="Test",
        data=[
            ScreenData(
                key="obj",
                example=DataSource(id="1", title="Example"),
            )
        ],
        layout=Layout(children=[]),
    )
    screen_2 = Screen(
        id="test",
        title="Test",
        data={
            "obj": {
                "type": "object",
                "properties": {
                    "id": {"type": "string"},
                    "title": {"type": "string"},
                },
                "__example__": {"id": "1", "title": "Example"},
            }
        },
        layout=Layout(children=[]),
    )

    flow_json = FlowJSON(screens=[screen_1, screen_2], version=3.0).to_dict()
    assert flow_json["screens"][0] == flow_json["screens"][1]

    # ---

    screen_1 = Screen(
        id="test",
        title="Test",
        data=[
            ScreenData(
                key="obj_array",
                example=[
                    DataSource(id="1", title="Example"),
                    DataSource(id="2", title="Example2"),
                ],
            ),
            ScreenData(
                key="str_array",
                example=["Example", "Example2"],
            ),
        ],
        layout=Layout(children=[]),
    )
    screen_2 = Screen(
        id="test",
        title="Test",
        data={
            "obj_array": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "id": {"type": "string"},
                        "title": {"type": "string"},
                    },
                },
                "__example__": [
                    {"id": "1", "title": "Example"},
                    {"id": "2", "title": "Example2"},
                ],
            },
            "str_array": {
                "type": "array",
                "items": {"type": "string"},
                "__example__": ["Example", "Example2"],
            },
        },
        layout=Layout(children=[]),
    )

    flow_json = FlowJSON(screens=[screen_1, screen_2], version=3.0).to_dict()
    assert flow_json["screens"][0] == flow_json["screens"][1]

    # ---

    with pytest.raises(ValueError):
        Screen(
            id="test",
            title="Test",
            data=[ScreenData(key="test", example=[])],
            layout=Layout(children=[]),
        )

    with pytest.raises(ValueError):
        Screen(
            id="test",
            title="Test",
            data=[ScreenData(key="test", example=ValueError)],
            layout=Layout(children=[]),
        )


def test_flow_response_with_error_msg():
    assert (
        "error_message"
        in FlowResponse(
            version=Version.FLOW_MSG.value,
            data={"test": "test"},
            screen="TEST",
            error_message="Example",
        ).to_dict()["data"]
    )


def test_flow_response_with_close_flow():
    assert (
        FlowResponse(  # closing flow make screen `SUCCESS`
            version=Version.FLOW_MSG.value,
            data={"test": "test"},
            close_flow=True,
            flow_token="test",
        ).to_dict()["screen"]
        == "SUCCESS"
    )

    with pytest.raises(ValueError):  # not closing flow without screen
        FlowResponse(
            version=Version.FLOW_MSG.value,
            data={"test": "test"},
            flow_token="test",
            close_flow=False,
        )

    with pytest.raises(ValueError):  # closing flow without flow_token
        FlowResponse(
            version=Version.FLOW_MSG.value,
            data={"test": "test"},
            close_flow=True,
        )

    with pytest.raises(ValueError):  # closing flow with error_message
        FlowResponse(
            version=Version.FLOW_MSG.value,
            data={"test": "test"},
            close_flow=True,
            flow_token="fdf",
            error_message="Example",
        )


def test_flow_response_with_data_sources():
    assert FlowResponse(
        version=Version.FLOW_MSG.value,
        data={"data_source": DataSource(id="1", title="Example")},
        screen="TEST",
    ).to_dict()["data"]["data_source"] == {"id": "1", "title": "Example"}

    assert FlowResponse(
        version=Version.FLOW_MSG.value,
        data={"data_source": [DataSource(id="1", title="Example")]},
        screen="TEST",
    ).to_dict()["data"]["data_source"] == [{"id": "1", "title": "Example"}]



"""This module contains the callback types."""

__all__ = [
    "CallbackButton",
    "CallbackSelection",
    "Button",
    "ButtonUrl",
    "SectionRow",
    "Section",
    "SectionList",
    "FlowButton",
    "CallbackData",
    "CallbackDataT",
]

import dataclasses
import datetime
import enum
import types
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Iterable,
    TypeVar,
    cast,
    get_origin,
    Union,
    get_args,
    Literal,
)

from .base_update import BaseUserUpdate  # noqa
from .flows import FlowStatus, FlowActionType
from .others import MessageType, Metadata, ReplyToMessage, User, InteractiveType
from .. import utils

if TYPE_CHECKING:
    from ..client import WhatsApp
    from .template import Template


class CallbackData:
    """
    Base class for all callback data classes. Subclass this class to create a type-safe callback data class.

        If you use ``dataclasses``, which is the recommended way (You get free ordered ``__init__`` and extra features),
        you should not use ``kw_only=True``.
        This is because we are limited to 200 characters in the callback data, so we need to use positional arguments.
        So object like ``User(id=123, name='John')`` will be converted to ``123:John``.

        Currently, the following types are supported:
        :class:`str`, :class:`int`, :class:`bool`, :class:`float` and :class:`enum.Enum` that
        inherits from :class:`str` (e.g ``class State(str, Enum)``). You can also use all the types as Optional or Union
        (e.g ``Optional[int]`` or ``Union[int, None]``). the Union length must be 2 at most, and one of the types must
        be ``None``. All fields can be with default values.

        The characters ``¶`` and ``~`` cannot be used when sending callbacks, because they are used as separators.
        You can change the separators by overriding ``__callback_data_sep__`` (``~`` for individual objects) and
        ``CallbackData.__callback_sep__`` (``¶`` in the base class level, affects all child classes).

        When providing subclassed ``CallbackData`` as a ``factory`` parameter in callback handlers, a basic matching
        filter (``startswith(callback_id)``) will be added automatically. So no need to create one yourself.


    Example:

        >>> from pywa.types import CallbackData
        >>> from dataclasses import dataclass # Use dataclass to get free ordered __init__
        >>> @dataclass(frozen=True, slots=True) # Do not use kw_only=True
        >>> class UserData(CallbackData): # Subclass CallbackData
        ...     id: int
        ...     name: str | None
        ...     admin: bool = False

        >>> from pywa import WhatsApp
        >>> from pywa.types import Button
        >>> wa = WhatsApp(...)
        >>> wa.send_message(
        ...     to='972987654321',
        ...     text='Click the button to get the user',
        ...     buttons=[
        ...         Button(
        ...             title='Get user',
        ...             callback_data=UserData(id=123, name='John', admin=True)
        ...         )
        ...     ]
        ... )

        >>> @wa.on_callback_button(factory=UserData) # Use the factory parameter to convert the callback data
        ... def on_user_data(client: WhatsApp, btn: CallbackButton[UserData]): # For autocomplete
        ...    if btn.data.admin: print(btn.data.id) # Access the data object as an attribute
    """

    __callback_id__: int = 0
    """Unique ID for each callback data class. Do not override this."""
    __callback_sep__: str = "¶"
    """The separator between multiple callback objects, Can be overridden globally. (Default ``¶``)"""
    __callback_data_sep__: str = "~"
    """The separator between the callback fields, Can be overridden individually. (Default ``~``)"""
    __callback_bool_true__: str = "§"
    """The string to represent ``True`` in the callback data. (Default ``§``)"""
    __callback_null__: str = "¤"
    """The string to represent ``None`` in the callback data. (Default ``¤``)"""
    __allowed_types__: tuple[type, ...] = (
        str,
        int,
        bool,
        float,
    )
    """The allowed types in the callback data."""

    def __init_subclass__(cls, *args, **kwargs):
        """Validate the callback data class and set a unique ID for it."""
        super().__init_subclass__(*args, **kwargs)
        if (
            len(
                set(
                    (
                        seps := {
                            "Callback Separator": cls.__callback_sep__,
                            "Data Separator": cls.__callback_data_sep__,
                            "Boolean True": cls.__callback_bool_true__,
                            "None Value": cls.__callback_null__,
                        }
                    ).values()
                )
            )
            != 4
        ):
            raise ValueError(
                f"Non-unique separators in callback data class `{cls.__name__}`: {seps}"
            )
        if not cls.__annotations__.items():
            raise TypeError(
                f"Callback data class `{cls.__name__}` must have at least one field."
            )
        unsupported_fields = set[tuple[str, type]]()
        for field_name, field_type in cls.__annotations__.items():
            if get_origin(field_type) in (types.UnionType, Union):
                if len(union_args := get_args(field_type)) > 2:
                    raise TypeError(
                        f"Field {field_name} in `{cls.__name__}` must be a Union of 2 types at most."
                    )
                if types.NoneType not in union_args:
                    raise TypeError(
                        f"Field `{field_name}` in `{cls.__name__}` must be an Optional[x] or Union[x, None]. "
                        f"(e.g. int | None)"
                    )
                field_type = next((a for a in union_args if a is not types.NoneType))
            if issubclass(field_type, enum.Enum) and not issubclass(field_type, str):
                raise TypeError(
                    f"Field `{field_name}` in `{cls.__name__}` must be an Enum that inherits from str."
                )
            if not issubclass(cast(type, field_type), cls.__allowed_types__):
                unsupported_fields.add((field_name, cast(type, field_type)))
        if unsupported_fields:
            raise TypeError(
                f"Unsupported types {unsupported_fields} in callback data. Use one of {cls.__allowed_types__}."
            )
        cls.__callback_id__ = CallbackData.__callback_id__
        CallbackData.__callback_id__ += 1

    @classmethod
    def from_str(
        cls,
        data: str,
    ) -> "CallbackData":
        """
        Internal function to convert a callback string to a callback object.
        """
        try:
            positional_args = []
            for annotation, value in zip(
                cls.__annotations__.values(),
                data.split(cls.__callback_data_sep__)[1:],
                strict=True,
            ):
                if get_origin(annotation) in (types.UnionType, Union):
                    if value == cls.__callback_null__:
                        positional_args.append(None)
                        continue
                    annotation = next(
                        a for a in get_args(annotation) if a is not types.NoneType
                    )
                if isinstance(annotation, bool):
                    if value == cls.__callback_bool_true__:
                        positional_args.append(True)
                        continue
                    elif value == "":
                        positional_args.append(False)
                        continue
                    raise ValueError
                positional_args.append(annotation(value))

            # noinspection PyArgumentList
            return cls(*positional_args)
        except (ValueError, TypeError) as e:
            raise ValueError(f"Invalid callback data for {cls.__name__}: {data}") from e

    @staticmethod
    def _not_contains(value: Any, *not_) -> str:
        """Internal function to validate that the value does not contain the separator."""
        if any(sep in (str_val := str(value)) for sep in not_):
            raise ValueError(
                f"Callback data cannot contain the characters {not_} "
                f"Because they are used as separators. \nYou can change the separators by overriding "
                f"``__callback_data_sep__`` (for individual objects) and ``CallbackData.__callback_sep__`` "
                f"(In the base class level, affects all child classes)."
            )
        return str_val

    def to_str(self) -> str:
        """
        Internal function to convert a callback object to a callback string.
        """
        values = [
            str(self.__callback_id__),
        ]
        for field_name, field_type in self.__annotations__.items():
            value = getattr(self, field_name)
            if value is None:
                value = self.__callback_null__
            if isinstance(value, bool):
                value = self.__callback_bool_true__ if value else ""
            if isinstance(value, enum.Enum):
                value = value.value
            values.append(self._not_contains(value, self.__callback_sep__))
        return self.__callback_data_sep__.join(values)

    @classmethod
    def join_to_str(cls, *datas: Any) -> str:
        """Internal function to join multiple callback objects to a callback string."""
        return cls.__callback_sep__.join(
            data.to_str()
            if isinstance(data, CallbackData)
            else cls._not_contains(data, cls.__callback_data_sep__)
            for data in datas
        )


CallbackDataT = TypeVar(
    "CallbackDataT",
    bound=str | CallbackData | Iterable[CallbackData | Any],
)
"""Type hint for ``callback_data`` parameter in :class:`Button` and :class:`SectionRow` and for ``tracker``'s"""


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class CallbackButton(BaseUserUpdate, Generic[CallbackDataT]):
    """
    Represents a callback button (Incoming update when user clicks on :class:`Button` or chooses
    :class:`Template.QuickReplyButtonData`).

    ``CallbackButton`` is a generic class, so when providing a ``factory`` parameter in callback handlers, you can
    specify the type of the factory to get autocomplete in the ``data`` attribute.

    Here is an example:

        >>> from pywa.types import CallbackData
        >>> from dataclasses import dataclass
        >>> @dataclass(frozen=True, slots=True)
        >>> class UserData(CallbackData): # Subclass CallbackData
        ...     id: int
        ...     name: str
        ...     admin: bool

        >>> from pywa import WhatsApp
        >>> from pywa.types import Button, CallbackButton
        >>> wa = WhatsApp(...)
        >>> wa.send_message(
        ...     to='972987654321',
        ...     text='Click the button to get the user',
        ...     buttons=[Button(title='Get user', callback_data=UserData(id=123, name='david', admin=True))]
        ... )                                     # Here ^^^ we use the UserData class as the callback data

        >>> @wa.on_callback_button(factory=UserData) # Use the factory parameter to convert the callback data
        ... def on_user_data(_: WhatsApp, btn: CallbackButton[UserData]): # For autocomplete
        ...    if btn.data.admin: print(btn.data.id) # Access the data object as an attribute

    You can even use multiple factories, and not only ``CallbackData`` subclasses!

        >>> from enum import Enum
        >>> class State(str, Enum):
        ...     START = 's'
        ...     END = 'e'

        >>> wa.send_message(
        ...     to='972987654321',
        ...     text='Click the button to get the user and state',
        ...     buttons=[Button(title='Get user', callback_data=(UserData(id=123, name='david', admin=True), State.START))]
        ... )                                     # Here ^^^ we send a tuple of UserData and State

        >>> @wa.on_callback_button(factory=(UserData, State)) # Use the factory parameter to convert the callback data
        ... def on_user_data(_: WhatsApp, btn: CallbackButton[tuple[UserData, State]]): # For autocomplete
        ...    user, state = btn.data # Unpack the tuple
        ...    if user.admin: print(user.id, state)


    Attributes:
        id: The ID of the message.
        metadata: The metadata of the message (to which phone number it was sent).
        type: The message type (:class:`MessageType.INTERACTIVE` for :class:`Button` presses or
        type: The message type (:class:`MessageType.INTERACTIVE` for :class:`Button` presses or
         :class:`MessageType.BUTTON` for :class:`Template.QuickReplyButtonData` choices).
        from_user: The user who sent the message.
        timestamp: The timestamp when the message was sent.
        reply_to_message: The message to which this callback button is a reply to.
        data: The data of the button (the ``callback_data`` parameter you provided in :class:`Button` or
         :class:`Template.QuickReplyButtonData`).
        title: The title of the button.
    """

    id: str
    type: MessageType
    metadata: Metadata
    from_user: User
    timestamp: datetime.datetime
    reply_to_message: ReplyToMessage
    data: CallbackDataT
    title: str

    _txt_fields = ("data",)

    @classmethod
    def from_update(cls, client: "WhatsApp", update: dict) -> "CallbackButton":
        msg = (value := update["entry"][0]["changes"][0]["value"])["messages"][0]
        match msg_type := msg["type"]:
            case MessageType.INTERACTIVE:
                title = msg["interactive"]["button_reply"]["title"]
                data = msg["interactive"]["button_reply"]["id"]
            case MessageType.BUTTON:
                title = msg["button"]["text"]
                data = msg["button"]["payload"]
            case _:
                raise ValueError(f"Invalid message type {msg_type}")
        return cls(
            _client=client,
            raw=update,
            id=msg["id"],
            metadata=Metadata.from_dict(value["metadata"]),
            type=MessageType(msg_type),
            from_user=User.from_dict(value["contacts"][0]),
            timestamp=datetime.datetime.fromtimestamp(int(msg["timestamp"])),
            reply_to_message=ReplyToMessage.from_dict(msg["context"]),
            data=data,
            title=title,
        )


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class CallbackSelection(BaseUserUpdate, Generic[CallbackDataT]):
    """
    Represents a callback selection (Incoming update when user clicks on :class:`SectionRow` in :class:`SectionList`).

    ``CallbackSelection`` is a generic class, so when providing a ``factory`` parameter in callback handlers, you can
    specify the type of the factory to get autocomplete in the ``data`` attribute.

    Here is an example:

        >>> from pywa.types import CallbackData
        >>> from dataclasses import dataclass
        >>> @dataclass(frozen=True, slots=True)
        >>> class UserData(CallbackData): # Subclass CallbackData
        ...     id: int
        ...     name: str
        ...     admin: bool

        >>> from pywa import WhatsApp
        >>> from pywa.types import SectionList, Section, SectionRow, CallbackSelection
        >>> wa = WhatsApp(...)
        >>> wa.send_message(
        ...     to='972987654321',
        ...     text='Click the button to get the user',
        ...     buttons=SectionList(
        ...         button_title='Get user', sections=[
        ...             Section(title='Users', rows=[
        ...                 SectionRow(title='Get user', callback_data=UserData(id=123, name='david', admin=True))
        ...             ])                              # Here ^^^ we use the UserData class as the callback data
        ...         ]
        ...     )
        ... )

        >>> @wa.on_callback_selection(factory=UserData) # Use the factory parameter to convert the callback data
        ... def on_user_data(_: WhatsApp, sel: CallbackSelection[UserData]): # For autocomplete
        ...    if sel.data.admin: print(sel.data.id) # Access the data object as an attribute

    You can even use multiple factories, and not only ``CallbackData`` subclasses!

        >>> from enum import Enum
        >>> class State(str, Enum):
        ...     START = 's'
        ...     END = 'e'

        >>> wa.send_message(
        ...     to='972987654321',
        ...     text='Click the button to get the user and state',
        ...     buttons=SectionList(
        ...         button_title='Get user', sections=[
        ...             Section(title='Users', rows=[
        ...                 SectionRow(title='Get user', callback_data=(UserData(id=123, name='david', admin=True), State.START))
        ...             ])                              # Here ^^^ we send a tuple of UserData and State
        ...         ]
        ...     )
        ... )

        >>> @wa.on_callback_selection(factory=(UserData, State)) # Use the factory parameter to convert the callback data
        ... def on_user_data(_: WhatsApp, sel: CallbackSelection[tuple[UserData, State]]): # For autocomplete
        ...    user, state = sel.data # Unpack the tuple
        ...    if user.admin: print(user.id, state)

    Attributes:
        id: The ID of the message.
        metadata: The metadata of the message (to which phone number it was sent).
        type: The message type (always :class:`MessageType.INTERACTIVE`).
        from_user: The user who sent the message.
        timestamp: The timestamp when the message was sent.
        reply_to_message: The message to which this callback selection is a reply to.
        data: The data of the selection (the ``callback_data`` parameter you provided in :class:`SectionRow`).
        title: The title of the selection.
        description: The description of the selection (optional).
    """

    id: str
    type: MessageType
    metadata: Metadata
    from_user: User
    timestamp: datetime.datetime
    reply_to_message: ReplyToMessage
    data: CallbackDataT
    title: str
    description: str | None

    _txt_fields = ("data",)

    @classmethod
    def from_update(cls, client: "WhatsApp", update: dict) -> "CallbackSelection":
        msg = (value := update["entry"][0]["changes"][0]["value"])["messages"][0]
        return cls(
            _client=client,
            raw=update,
            id=msg["id"],
            metadata=Metadata.from_dict(value["metadata"]),
            type=MessageType(msg["type"]),
            from_user=User.from_dict(value["contacts"][0]),
            timestamp=datetime.datetime.fromtimestamp(int(msg["timestamp"])),
            reply_to_message=ReplyToMessage.from_dict(msg["context"]),
            data=msg["interactive"]["list_reply"]["id"],
            title=msg["interactive"]["list_reply"]["title"],
            description=msg["interactive"]["list_reply"].get("description"),
        )


def _resolve_callback_data(data: CallbackDataT) -> str:
    """Internal function to convert callback data to a string."""
    if isinstance(data, str):
        return data
    elif isinstance(data, CallbackData):
        return data.to_str()
    elif isinstance(data, Iterable):
        if any(isinstance(item, CallbackData) for item in data):
            return CallbackData.join_to_str(*data)
    raise TypeError(f"Invalid callback data type {type(data)}")


@dataclasses.dataclass(slots=True)
class Button:
    """
    Represents a button in the button list.

    Attributes:
        title: The title of the button (up to 20 characters).
        callback_data: The data to send when the user clicks on the button (up to 256 characters, for complex data
         You can use :class:`CallbackData`).
    """

    title: str
    callback_data: CallbackDataT

    def to_dict(self) -> dict:
        return {
            "type": "reply",
            "reply": {
                "id": _resolve_callback_data(self.callback_data),
                "title": self.title,
            },
        }


@dataclasses.dataclass(slots=True)
class ButtonUrl:
    """
    Represents a button in the bottom of the message that opens a URL.

    Attributes:
        title: The title of the button (up to 20 characters).
        url: The URL to open when the user clicks on the button.
    """

    title: str
    url: str

    def to_dict(self) -> dict:
        return {
            "name": InteractiveType.CTA_URL,
            "parameters": {"display_text": self.title, "url": self.url},
        }


@dataclasses.dataclass(slots=True)
class SectionRow:
    """
    Represents a choice row in a section rows.

    Attributes:
        title: The title of the row (up to 24 characters).
        callback_data: The payload to send when the user clicks on the row (up to 200 characters, for complex data
            You can use :class:`CallbackData`).
        description: The description of the row (optional, up to 72 characters).
    """

    title: str
    callback_data: CallbackDataT
    description: str | None = None

    def to_dict(self) -> dict:
        d = {
            "id": _resolve_callback_data(self.callback_data),
            "title": self.title,
        }
        if self.description:
            d["description"] = self.description
        return d


@dataclasses.dataclass(slots=True)
class Section:
    """
    Represents a section in a section list.

    Attributes:
        title: The title of the section (up to 24 characters).
        rows: The rows in the section (at least 1, no more than 10).
    """

    title: str
    rows: Iterable[SectionRow]

    def to_dict(self) -> dict:
        return {
            "title": self.title,
            "rows": tuple(row.to_dict() for row in self.rows),
        }


@dataclasses.dataclass(slots=True)
class SectionList:
    """
    Represents a section list in an interactive message.

    Attributes:
        button_title: The title of the button that opens the section list (up to 20 characters).
        sections: The sections in the section list (at least 1, no more than 10).
    """

    button_title: str
    sections: Iterable[Section]

    def to_dict(self) -> dict:
        return {
            "button": self.button_title,
            "sections": tuple(section.to_dict() for section in self.sections),
        }


@dataclasses.dataclass(slots=True)
class FlowButton:
    """
    Represents a button that opens a flow.

    Attributes:
        title: Text on the CTA button. e.g ``SignUp``, Up to 20 characters, no emojis)
        flow_id: Unique ID of the Flow provided by WhatsApp.
        flow_token: Flow token generated by the business to serve as an identifier for data exchange.
        flow_message_version: Version of the flow message. Default is the latest version.
        flow_action_type: Type of action to be performed when the user clicks on the CTA button.
        flow_action_screen: The ID of the first Screen.
         Required when ``flow_action_type`` is ``FlowActionType.NAVIGATE`` (default).
        flow_action_payload: The payload to send when the user clicks on the button
         (optional, only when ``flow_action_type`` is ``FlowActionType.NAVIGATE``).
        mode: The mode of the flow. ``FlowStatus.PUBLISHED`` (default) or ``FlowStatus.DRAFT`` (for testing).
    """

    title: str
    flow_id: str | int
    flow_token: str
    flow_action_type: (
        Literal[FlowActionType.NAVIGATE, FlowActionType.DATA_EXCHANGE] | None
    ) = None
    flow_action_screen: str | None = None
    flow_action_payload: dict[str, Any] | None = None
    flow_message_version: int | float | str | Literal[utils.Version.FLOW_MSG] = (
        utils.Version.FLOW_MSG
    )
    mode: Literal[FlowStatus.PUBLISHED, FlowStatus.DRAFT] = FlowStatus.PUBLISHED

    def __post_init__(self):
        utils.Version.FLOW_MSG.validate_min_version(str(self.flow_message_version))
        if (
            self.flow_action_type == FlowActionType.NAVIGATE
            and self.flow_action_screen is None
        ):
            raise ValueError(
                "flow_action_screen cannot be None when flow_action_type is FlowActionType.NAVIGATE"
            )

    def to_dict(self) -> dict:
        return {
            "name": InteractiveType.FLOW,
            "parameters": {
                "mode": self.mode.lower(),
                "flow_message_version": str(self.flow_message_version),
                "flow_token": self.flow_token,
                "flow_id": self.flow_id,
                "flow_cta": self.title,
                **(
                    {"flow_action": str(self.flow_action_type)}
                    if self.flow_action_type is not None
                    else {}
                ),
                **(
                    {
                        "flow_action_payload": {
                            "screen": self.flow_action_screen,
                            **(
                                {"data": self.flow_action_payload}
                                if self.flow_action_payload is not None
                                else {}
                            ),
                        }
                    }
                    if self.flow_action_type == FlowActionType.NAVIGATE
                    else {}
                ),
            },
        }



import dataclasses

import pytest

from pywa.types import CallbackData


def test_callback_class_uniqueness():
    """Test the uniqueness of the callback data."""

    @dataclasses.dataclass(slots=True, frozen=True)
    class User(CallbackData):
        id: str
        name: str
        is_admin: bool

    @dataclasses.dataclass(slots=True, frozen=True)
    class Group(CallbackData):
        id: str
        name: str

    assert (
        User.__callback_id__ != Group.__callback_id__
    ), "The callback id must be unique for each child class."


def test_callback_class_not_empty():
    """Test if the callback data is empty."""
    with pytest.raises(TypeError):

        @dataclasses.dataclass(slots=True, frozen=True)
        class User1(CallbackData):
            pass

    with pytest.raises(TypeError):

        @dataclasses.dataclass(slots=True, frozen=True)
        class User2(CallbackData):
            id: str
            name: str
            is_admin: bool
            x: int

        @dataclasses.dataclass(slots=True, frozen=True)
        class UserChild(User2):
            pass


def test_callback_supported_types():
    """Test the supported types of the callback data."""
    with pytest.raises(TypeError):

        @dataclasses.dataclass(slots=True, frozen=True)
        class User(CallbackData):
            id: int
            name: str
            is_admin: bool
            x: dict
            y: list
            z: set


def test_data():
    """Test the callback data."""

    @dataclasses.dataclass(slots=True, frozen=True)
    class User(CallbackData):
        id: int
        name: str
        is_admin: bool

    assert User(id=1234, name="xxx", is_admin=True).to_str() == "7~1234~xxx~§"
    assert User(id=3456, name="yyy", is_admin=False).to_str() == "7~3456~yyy~"
    assert User.from_str("1~1234~xxx~§") == User(id=1234, name="xxx", is_admin=True)
    assert User.from_str("1~3456~yyy~") == User(id=3456, name="yyy", is_admin=False)

    with pytest.raises(ValueError):  # too few fields
        User.from_str("1~3456")

    with pytest.raises(ValueError):  # too many fields
        User.from_str("1~3456~yyy~§~")

    with pytest.raises(ValueError):  # invalid type
        User.from_str("1~x~yyy~§")


def test_multiple_data():
    """Test multiple callback data."""

    @dataclasses.dataclass(slots=True, frozen=True)
    class User(CallbackData):
        id: str
        name: str
        is_admin: bool

    @dataclasses.dataclass(slots=True, frozen=True)
    class Group(CallbackData):
        id: str
        name: str

    user = User(id="1234", name="xxx", is_admin=True)
    group = Group(id="3456", name="yyy")
    assert CallbackData.join_to_str(user, group) == "9~1234~xxx~§¶11~3456~yyy"
    assert (
        CallbackData.join_to_str(user, group, user)
        == "9~1234~xxx~§¶11~3456~yyy¶9~1234~xxx~§"
    )
    x, y = "3~1234~xxx~§¶5~3456~yyy".split(CallbackData.__callback_sep__)
    assert (User.from_str(x), Group.from_str(y)) == (user, group)
    x, y, z = "3~1234~xxx~§¶5~3456~yyy¶3~1234~xxx~§".split(
        CallbackData.__callback_sep__
    )
    assert (User.from_str(x), Group.from_str(y), User.from_str(z)) == (
        user,
        group,
        user,
    )


def test_data_sep():
    """Test the data separator override."""

    @dataclasses.dataclass(slots=True, frozen=True)
    class User(CallbackData):
        __callback_data_sep__ = "*"
        id: int
        name: str
        is_admin: bool

    assert User(id=1234, name="xxx", is_admin=True).to_str() == "13*1234*xxx*§"
    assert User(id=3456, name="yyy", is_admin=False).to_str() == "13*3456*yyy*"

    with pytest.raises(ValueError):
        User.from_str("1*3456*David*Lev*")

    try:
        User.from_str("1*3456*David~Lev*")
    except ValueError:
        pytest.fail("The data separator override does not work.")


def test_data_sep_multiple_data():
    """Test the data separator override with multiple data."""
    CallbackData.__callback_sep__ = "^"

    @dataclasses.dataclass(slots=True, frozen=True)
    class User(CallbackData):
        id: int
        name: str
        is_admin: bool

    @dataclasses.dataclass(slots=True, frozen=True)
    class Group(CallbackData):
        id: int
        name: str

    user = User(id=1234, name="xxx", is_admin=True)
    group = Group(id=3456, name="yyy")
    assert CallbackData.join_to_str(user, group) == "15~1234~xxx~§^17~3456~yyy"
    assert (
        CallbackData.join_to_str(user, group, user)
        == "15~1234~xxx~§^17~3456~yyy^15~1234~xxx~§"
    )
    x, y = "3~1234~xxx~§^5~3456~yyy".split(CallbackData.__callback_sep__)
    assert (User.from_str(x), Group.from_str(y)) == (user, group)
    x, y, z = "3~1234~xxx~§^5~3456~yyy^3~1234~xxx~§".split(
        CallbackData.__callback_sep__
    )
    assert (User.from_str(x), Group.from_str(y), User.from_str(z)) == (
        user,
        group,
        user,
    )



from __future__ import annotations

from types import MappingProxyType

"""This module contains the types related to messages."""

__all__ = ["Message"]

import dataclasses
import datetime
from typing import TYPE_CHECKING, Any, Callable, Iterable

from ..errors import WhatsAppError

from .base_update import BaseUserUpdate  # noqa
from .callback import Button, ButtonUrl, SectionList
from .media import Audio, Document, Image, Sticker, Video
from .others import (
    Contact,
    Location,
    MessageType,
    Metadata,
    Order,
    ProductsSection,
    Reaction,
    ReplyToMessage,
    System,
    User,
)

if TYPE_CHECKING:
    from ..client import WhatsApp


_FIELDS_TO_OBJECTS_CONSTRUCTORS: dict[str, Callable[[dict, WhatsApp], Any]] = dict(
    text=lambda m, _client: m["body"],
    image=Image.from_dict,
    video=Video.from_dict,
    sticker=Sticker.from_dict,
    document=Document.from_dict,
    audio=Audio.from_dict,
    reaction=Reaction.from_dict,
    location=Location.from_dict,
    contacts=lambda m, _client: tuple(Contact.from_dict(c) for c in m),
    order=Order.from_dict,
    system=System.from_dict,
)


@dataclasses.dataclass(frozen=True, slots=True, kw_only=True)
class Message(BaseUserUpdate):
    """
    A message received from a user.

    - `'Message' on developers.facebook.com <https://developers.facebook.com/docs/whatsapp/cloud-api/webhooks/components#messages-object>`_

    Attributes:
        id: The message ID (If you want to reply to the message, use ``message_id_to_reply`` instead).
        metadata: The metadata of the message (to which phone number it was sent).
        type: The message type (See :class:`MessageType`).
        from_user: The user who sent the message.
        timestamp: The timestamp when the message was sent.
        reply_to_message: The message to which this message is a reply to. (Optional)
        forwarded: Whether the message was forwarded.
        forwarded_many_times: Whether the message was forwarded many times.
         (when True, ``forwarded`` will be True as well)
        text: The text of the message (if the message type is :class:`MessageType.TEXT`).
        image: The image of the message (if the message type is :class:`MessageType.IMAGE`).
        video: The video of the message (if the message type is :class:`MessageType.VIDEO`).
        sticker: The sticker of the message (if the message type is :class:`MessageType.STICKER`).
        document: The document of the message (if the message type is :class:`MessageType.DOCUMENT`).
        audio: The audio of the message (if the message type is :class:`MessageType.AUDIO`).
        caption: The caption of the message (Optional, only available for :class:`MessageType.IMAGE`,
         :class:`MessageType.VIDEO` and :class:`MessageType.DOCUMENT`).
        reaction: The reaction of the message (if the message type is :class:`MessageType.REACTION`).
        location: The location of the message (if the message type is :class:`MessageType.LOCATION`).
        contacts: The contacts of the message (if the message type is :class:`MessageType.CONTACTS`).
        order: The order of the message (if the message type is :class:`MessageType.ORDER`).
        system: The system update (if the message type is :class:`MessageType.SYSTEM`).
        error: The error of the message (if the message type is :class:`MessageType.UNSUPPORTED`).
    """

    id: str
    type: MessageType
    metadata: Metadata
    from_user: User
    timestamp: datetime.datetime
    reply_to_message: ReplyToMessage | None
    forwarded: bool
    forwarded_many_times: bool
    text: str | None = None
    image: Image | None = None
    video: Video | None = None
    sticker: Sticker | None = None
    document: Document | None = None
    audio: Audio | None = None
    caption: str | None = None
    reaction: Reaction | None = None
    location: Location | None = None
    contacts: tuple[Contact, ...] | None = None
    order: Order | None = None
    system: System | None = None
    error: WhatsAppError | None = None

    _media_fields = {"image", "video", "sticker", "document", "audio"}
    _txt_fields = ("text", "caption")
    _fields_to_objects_constructors = MappingProxyType(
        dict(
            text=lambda m, _client: m["body"],
            image=Image.from_dict,
            video=Video.from_dict,
            sticker=Sticker.from_dict,
            document=Document.from_dict,
            audio=Audio.from_dict,
            reaction=Reaction.from_dict,
            location=Location.from_dict,
            contacts=lambda m, _client: tuple(Contact.from_dict(c) for c in m),
            order=Order.from_dict,
            system=System.from_dict,
        )
    )
    """A mapping of message types to their respective constructors."""

    @property
    def message_id_to_reply(self) -> str:
        """The ID of the message to reply to."""
        return (
            self.id if self.type != MessageType.REACTION else self.reaction.message_id
        )

    @property
    def has_media(self) -> bool:
        """
        Whether the message has any media. (image, video, sticker, document or audio)
            - If you want to get the media of the message, use :attr:`~Message.media` instead.
        """
        return self.media is not None

    @property
    def is_reply(self) -> bool:
        """
        Whether the message is a reply to another message.

        - Reaction messages are also considered as replies (But ``.reply_to_message`` will be ``None``).
        """
        return self.reply_to_message is not None or self.reaction is not None

    @classmethod
    def from_update(cls, client: WhatsApp, update: dict) -> Message:
        msg = (value := update["entry"][0]["changes"][0]["value"])["messages"][0]
        error = value.get("errors", msg.get("errors", (None,)))[0]
        msg_type = msg["type"]
        context = msg.get("context", {})
        constructor = cls._fields_to_objects_constructors.get(msg_type)
        # noinspection PyArgumentList
        msg_content = (
            {msg_type: constructor(msg[msg_type], _client=client)}
            if constructor is not None
            else {}
        )
        try:
            usr = User.from_dict(value["contacts"][0])
        except KeyError:
            usr = User(
                wa_id=msg["from"], name=None
            )  # some messages don't have contacts
        return cls(
            _client=client,
            raw=update,
            id=msg["id"],
            type=MessageType(msg_type),
            **msg_content,
            from_user=usr,
            timestamp=datetime.datetime.fromtimestamp(int(msg["timestamp"])),
            metadata=Metadata.from_dict(value["metadata"]),
            forwarded=context.get("forwarded", False)
            or context.get("frequently_forwarded", False),
            forwarded_many_times=context.get("frequently_forwarded", False),
            reply_to_message=ReplyToMessage.from_dict(context),
            caption=msg.get(msg_type, {}).get("caption")
            if msg_type in cls._media_fields
            else None,
            error=WhatsAppError.from_dict(error=error) if error is not None else None,
        )

    @property
    def media(
        self,
    ) -> Image | Video | Sticker | Document | Audio | None:
        """
        The media of the message if any, otherwise ``None``. (image, video, sticker, document or audio)
            - If you want to check whether the message has any media, use :attr:`~Message.has_media` instead.
        """
        return next(
            (
                getattr(self, media_type)
                for media_type in self._media_fields
                if getattr(self, media_type)
            ),
            None,
        )

    def download_media(
        self,
        filepath: str | None = None,
        filename: str | None = None,
        in_memory: bool = False,
        **kwargs,
    ) -> str | bytes:
        """
        Download a media file from WhatsApp servers (image, video, sticker, document or audio).

        Args:
            filepath: The path where to save the file (if not provided, the current working directory will be used).
            filename: The name of the file (if not provided, it will be guessed from the URL + extension).
            in_memory: Whether to return the file as bytes instead of saving it to disk (default: False).
            **kwargs: Additional arguments to pass to requests.get.

        Returns:
            The path of the saved file if ``in_memory`` is False, the file as bytes otherwise.

        Raises:
            ValueError: If the message does not contain any media.
        """
        try:
            return self.media.download(
                path=filepath, filename=filename, in_memory=in_memory, **kwargs
            )
        except AttributeError:
            raise ValueError("Message does not contain any media.")

    def copy(
        self,
        to: str,
        header: str | None = None,
        body: str | None = None,
        footer: str | None = None,
        buttons: Iterable[Button] | ButtonUrl | SectionList | None = None,
        preview_url: bool = False,
        reply_to_message_id: str = None,
        keyboard: None = None,
        tracker: str | None = None,
        sender: str | int | None = None,
    ) -> str:
        """
        Send the message to another user.

            - The WhatsApp Cloud API does not offer a `real` forward option, so this method will send a new message with the same content as the original message.
            - Supported message types: ``TEXT``, ``DOCUMENT``, ``IMAGE``, ``VIDEO``, ``STICKER``, ``LOCATION``, ``AUDIO``, ``CONTACTS``, ``ORDER`` and ``SYSTEM``.
            - If the message type is ``reaction``, you must provide ``reply_to_message_id``.

        Args:
            to: The phone ID of the WhatsApp user to copy the message to.
            header: The header of the message (if keyboard is provided, optional, up to 60 characters, no markdown allowed).
            body: The body of the message (if keyboard are provided, optional, up to 1024 characters, markdown allowed).
            footer: The footer of the message (if keyboard is provided, optional, markdown has no effect).
            buttons: The buttons to send with the message (only in case of message from type ``text``, ``document``,
             ``video`` and ``image``. also, the ``SectionList`` is only available to ``text`` type)
            reply_to_message_id:  The message ID to reply to (optional).
            preview_url: Whether to show a preview of the URL in the message (if any).
            keyboard: Deprecated and will be removed in a future version, use ``buttons`` instead.
            tracker: The track data of the message.
            sender: The phone ID to send the message from (optional, overrides the client's phone ID).

        Returns:
            The ID of the sent message.

        Raises:
            ValueError: If the message type is ``reaction`` and no ``reply_to_message_id`` is provided, or if the message
             type is ``unsupported``.
        """
        match self.type:
            case MessageType.TEXT:
                return self._client.send_message(
                    sender=sender,
                    to=to,
                    text=self.text,
                    preview_url=preview_url,
                    header=header,
                    footer=footer,
                    buttons=buttons,
                    reply_to_message_id=reply_to_message_id,
                    keyboard=keyboard,
                    tracker=tracker,
                )
            case MessageType.DOCUMENT:
                return self._client.send_document(
                    sender=sender,
                    to=to,
                    document=self.document.id,
                    filename=self.document.filename,
                    caption=self.caption,
                    body=body,
                    footer=footer,
                    buttons=keyboard or buttons,
                    reply_to_message_id=reply_to_message_id,
                    tracker=tracker,
                )
            case MessageType.IMAGE:
                return self._client.send_image(
                    sender=sender,
                    to=to,
                    image=self.image.id,
                    caption=self.caption,
                    body=body,
                    footer=footer,
                    buttons=keyboard or buttons,
                    reply_to_message_id=reply_to_message_id,
                    tracker=tracker,
                )
            case MessageType.VIDEO:
                return self._client.send_video(
                    sender=sender,
                    to=to,
                    video=self.video.id,
                    caption=self.caption,
                    buttons=keyboard or buttons,
                    body=body,
                    footer=footer,
                    reply_to_message_id=reply_to_message_id,
                    tracker=tracker,
                )
            case MessageType.STICKER:
                return self._client.send_sticker(
                    sender=sender, to=to, sticker=self.sticker.id, tracker=tracker
                )
            case MessageType.LOCATION:
                return self._client.send_location(
                    sender=sender,
                    to=to,
                    latitude=self.location.latitude,
                    longitude=self.location.longitude,
                    name=self.location.name,
                    address=self.location.address,
                    tracker=tracker,
                )
            case MessageType.AUDIO:
                return self._client.send_audio(
                    sender=sender, to=to, audio=self.audio.id, tracker=tracker
                )
            case MessageType.CONTACTS:
                return self._client.send_contact(
                    sender=sender,
                    to=to,
                    contact=self.contacts,
                    reply_to_message_id=reply_to_message_id,
                    tracker=tracker,
                )
            case MessageType.REACTION:
                if reply_to_message_id is None:
                    raise ValueError(
                        "You need to provide `reply_to_message_id` in order to `copy` a reaction"
                    )
                return self._client.send_reaction(
                    sender=sender,
                    to=to,
                    message_id=reply_to_message_id,
                    emoji=self.reaction.emoji or "",
                )
            case MessageType.ORDER:
                if len(self.order.products) == 1:
                    return self._client.send_product(
                        sender=sender,
                        to=to,
                        catalog_id=self.order.catalog_id,
                        sku=self.order.products[0].sku,
                        body=body,
                        footer=footer,
                        reply_to_message_id=reply_to_message_id,
                    )
                return self._client.send_products(
                    sender=sender,
                    to=to,
                    catalog_id=self.order.catalog_id,
                    product_sections=(
                        ProductsSection(
                            title=header,
                            skus=(p.sku for p in self.order.products),
                        ),
                    ),
                    title=header,
                    body=body,
                    footer=footer,
                    reply_to_message_id=reply_to_message_id,
                )
            case MessageType.SYSTEM:
                return self._client.send_message(
                    sender=sender,
                    to=to,
                    text=self.system.body,
                    header=header,
                    footer=footer,
                    buttons=keyboard,
                    reply_to_message_id=reply_to_message_id,
                )
            case _:
                raise ValueError(f"Message of type {self.type} cannot be copied.")



from __future__ import annotations

from .flows import FlowActionType

"""This module contains the types related to templates."""

__all__ = [
    "Template",
    "NewTemplate",
    "TemplateResponse",
    "TemplateStatus",
]

import abc
import dataclasses
import logging
import re
import pathlib
import datetime
from typing import TYPE_CHECKING, Any, BinaryIO, Iterable, Literal

from .. import utils

from .base_update import BaseUpdate  # noqa
from .callback import CallbackDataT, _resolve_callback_data  # noqa
from .others import ProductsSection

if TYPE_CHECKING:
    from ..client import WhatsApp

_logger = logging.getLogger(__name__)


def _get_examples_from_placeholders(
    string: str, start: str = "{", end: str = "}"
) -> tuple[str, tuple[str, ...]]:
    """
    Extracts the examples from a string.

    Example:

        >>> _get_examples_from_placeholders('Hello, {john}, today is {day}')
        ('Hello, {{1}}, today is {{2}}', ('john', 'day'))
        >>> _get_examples_from_placeholders('Hello, (john), today is (day)',start='(',end=')')

    Args:
      string: The string to extract the examples from.
      start: The start of the example (default: ``'{'``).
      end: The end of the example (default: ``'}'``).

    Returns:
      A tuple of the formatted string and the examples.
    """
    examples: list[str] = []
    for example in re.finditer(r"\%s(.*?)\%s" % (start, end), string):
        examples.append(example.group(1))
    for idx, example in enumerate(examples, start=1):
        string = string.replace(f"{start}{example}{end}", "{{" + str(idx) + "}}")
    return string, tuple(examples)


@dataclasses.dataclass(frozen=True, slots=True)
class TemplateResponse(utils.FromDict):
    """

    Attributes:
        id: the template ID.
        status: the template status.
        category: the template category.
    """

    id: str
    status: str
    category: NewTemplate.Category


class ComponentType(utils.StrEnum):
    HEADER = "HEADER"
    BODY = "BODY"
    FOOTER = "FOOTER"
    BUTTONS = "BUTTONS"


class HeaderFormatType(utils.StrEnum):
    TEXT = "TEXT"
    IMAGE = "IMAGE"
    VIDEO = "VIDEO"
    DOCUMENT = "DOCUMENT"
    LOCATION = "LOCATION"


class ButtonType(utils.StrEnum):
    PHONE_NUMBER = "PHONE_NUMBER"
    URL = "URL"
    QUICK_REPLY = "QUICK_REPLY"
    OTP = "OTP"
    MPM = "MPM"
    CATALOG = "CATALOG"
    COPY_CODE = "COPY_CODE"
    FLOW = "FLOW"


class Language(utils.StrEnum):
    """
    Template language and locale code.
        - Both :class:`NewTemplate` and :class:`Template` have this class as an attribute.

    >>> from pywa.types import NewTemplate
    >>> NewTemplate.Language.ENGLISH_US

    >>> from pywa.types import Template
    >>> Template.Language.ENGLISH_US

    `'Template language and locale code' on
    developers.facebook.com
    <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates/supported-languages>`_
    """

    AFRIKAANS = "af"
    ALBANIAN = "sq"
    ARABIC = "ar"
    AZERBAIJANI = "az"
    BENGALI = "bn"
    BULGARIAN = "bg"
    CATALAN = "ca"
    CHINESE_CHN = "zh_CN"
    CHINESE_HKG = "zh_HK"
    CHINESE_TAI = "zh_TW"
    CROATIAN = "hr"
    CZECH = "cs"
    DANISH = "da"
    DUTCH = "nl"
    ENGLISH = "en"
    ENGLISH_UK = "en_GB"
    ENGLISH_US = "en_US"
    ESTONIAN = "et"
    FILIPINO = "fil"
    FINNISH = "fi"
    FRENCH = "fr"
    GEORGIAN = "ka"
    GERMAN = "de"
    GREEK = "el"
    GUJARATI = "gu"
    HAUSA = "ha"
    HEBREW = "he"
    HINDI = "hi"
    HUNGARIAN = "hu"
    INDONESIAN = "id"
    IRISH = "ga"
    ITALIAN = "it"
    JAPANESE = "ja"
    KANNADA = "kn"
    KAZAKH = "kk"
    KINYARWANDA = "rw_RW"
    KOREAN = "ko"
    KYRGYZ_KG = "ky_KG"
    LAO = "lo"
    LATVIAN = "lv"
    LITHUANIAN = "lt"
    MACEDONIAN = "mk"
    MALAY = "ms"
    MALAYALAM = "ml"
    MARATHI = "mr"
    NORWEGIAN = "nb"
    PERSIAN = "fa"
    POLISH = "pl"
    PORTUGUESE_BR = "pt_BR"
    PORTUGUESE_POR = "pt_PT"
    PUNJABI = "pa"
    ROMANIAN = "ro"
    RUSSIAN = "ru"
    SERBIAN = "sr"
    SLOVAK = "sk"
    SLOVENIAN = "sl"
    SPANISH = "es"
    SPANISH_ARG = "es_AR"
    SPANISH_SPA = "es_ES"
    SPANISH_MEX = "es_MX"
    SWAHILI = "sw"
    SWEDISH = "sv"
    TAMIL = "ta"
    TELUGU = "te"
    THAI = "th"
    TURKISH = "tr"
    UKRAINIAN = "uk"
    URDU = "ur"
    UZBEK = "uz"
    VIETNAMESE = "vi"
    ZULU = "zu"


class NewTemplateComponentABC(abc.ABC):
    @property
    @abc.abstractmethod
    def type(self) -> ComponentType: ...


class NewTemplateHeaderABC(NewTemplateComponentABC, abc.ABC):
    @property
    @abc.abstractmethod
    def format(self) -> HeaderFormatType: ...


class NewButtonABC(abc.ABC):
    @property
    @abc.abstractmethod
    def type(self) -> ButtonType: ...

    @abc.abstractmethod
    def to_dict(self, placeholder: tuple[str, str] = None) -> dict[str, str | None]: ...


@dataclasses.dataclass(slots=True)
class NewTemplate:
    """
    Represents a new template.
    `'Create Templates' on developers.facebook.com
    <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates>`_.

    Attributes:
        name: Name of the template (up to 512 characters, must be unique).
        category: Category of the template.
        language: The language of the template (See `Template language and locale code
         <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates/supported-languages>`_).
        body: Body of the template.
        header: Header of the template (optional).
        footer: Footer of the template (optional).
        buttons: Buttons to send with the template (optional, ``OTPButton`` is required for AUTHENTICATION).

    Example:

        >>> from pywa.types import NewTemplate as NewTemp
        >>> NewTemp(
        ...     name='buy_new_iphone_x',
        ...     category=NewTemp.Category.MARKETING,
        ...     language=NewTemp.Language.ENGLISH_US,
        ...     header=NewTemp.Text('The New iPhone {15} is here!'),
        ...     body=NewTemp.Body('Hello {John}! Buy now and use the code {WA_IPHONE_15} to get {15%} off!'),
        ...     footer=NewTemp.Footer('Powered by PyWa'),
        ...     buttons=[
        ...         NewTemp.UrlButton(title='Buy Now', url='https://example.com/shop/{iphone15}'),
        ...         NewTemp.PhoneNumberButton(title='Call Us', phone_number='1234567890'),
        ...         NewTemp.QuickReplyButton('Unsubscribe from marketing messages'),
        ...         NewTemp.QuickReplyButton('Unsubscribe from all messages'),
        ...     ],
        ... )

        Example for Authentication template:

        >>> from pywa.types import NewTemplate as NewTemp
        >>> NewTemp(
        ...     name='auth_with_otp',
        ...     category=NewTemp.Category.AUTHENTICATION,
        ...     language=NewTemp.Language.ENGLISH_US,
        ...     body=NewTemp.AuthBody(
        ...         code_expiration_minutes=5,
        ...         add_security_recommendation=True
        ...     ),
        ...     buttons=NewTemp.OTPButton(
        ...         otp_type=NewTemp.OTPButton.OtpType.ONE_TAP,
        ...         title='Copy Code',
        ...         autofill_text='Autofill',
        ...         package_name='com.example.app',
        ...         signature_hash='1234567890ABCDEF1234567890ABCDEF12345678'
        ...     )
        ... )

    Templates are limited to 10 quick reply buttons. If using quick reply buttons with other buttons, buttons must be
    organized into two groups: quick reply buttons and non-quick reply buttons. If grouped incorrectly, error will be
    raised when submitting the template.

    When you send a template that has multiple quick reply buttons, the order in which the buttons appear in the template
    is the order in which they will appear in the delivered message
    """

    name: str
    category: Category
    language: Language | str
    body: Body | AuthBody
    header: Text | Image | Video | Document | Location | None = None
    footer: Footer | None = None
    buttons: (
        Iterable[PhoneNumberButton | UrlButton | QuickReplyButton | CopyCodeButton]
        | MPMButton
        | CatalogButton
        | OTPButton
        | FlowButton
        | None
    ) = None

    def __post_init__(self):
        if self.category == self.Category.AUTHENTICATION and not (
            isinstance(self.body, self.AuthBody)
            or isinstance(self.buttons, self.OTPButton)
        ):
            raise ValueError(
                "body of AuthBody and buttons of OTPButton are required for AUTHENTICATION"
            )

    def to_dict(self, placeholder: tuple[str, str] = None) -> dict[str, Any]:
        if isinstance(self.buttons, self.OTPButton):
            components = (
                dict(
                    type=ComponentType.BUTTONS.value,
                    buttons=(self.buttons.to_dict(),),
                ),
                dict(
                    type=ComponentType.BODY.value,
                    add_security_recommendation=self.body.add_security_recommendation,
                ),
                dict(
                    type=ComponentType.FOOTER.value,
                    code_expiration_minutes=self.body.code_expiration_minutes,
                )
                if self.body.code_expiration_minutes
                else None,
            )
        else:
            components = (
                self.body.to_dict(placeholder),
                self.header.to_dict(placeholder) if self.header else None,
                self.footer.to_dict() if self.footer else None,
                dict(
                    type=ComponentType.BUTTONS.value,
                    buttons=tuple(
                        button.to_dict(placeholder)
                        for button in (
                            self.buttons
                            if isinstance(self.buttons, Iterable)
                            else (self.buttons,)
                        )
                    ),
                )
                if self.buttons
                else None,
            )
        return dict(
            name=self.name,
            category=self.category.value,
            language=str(self.language),
            components=tuple(
                component for component in components if component is not None
            ),
        )

    Language = Language

    class Category(utils.StrEnum):
        """
        Template category.

        `'Template Categorization' on
        developers.facebook.com
        <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates#categories>`_

        Attributes:
            AUTHENTICATION: Authentication templates are used to send one-time passwords (OTPs) or codes to app users.
            MARKETING: Marketing templates are used to send promotional messages to app users.
            UTILITY: Utility templates are used to send non-promotional messages to app users.
        """

        AUTHENTICATION = "AUTHENTICATION"
        MARKETING = "MARKETING"
        UTILITY = "UTILITY"

    @dataclasses.dataclass(slots=True)
    class Text(NewTemplateHeaderABC):
        """
        Represents a text header.

        Example:

            >>> from pywa.types import NewTemplate
            >>> NewTemplate.Text(text='Hello, {John}!')

        Attributes:
            text: Text to send with the header (Up to 60 characters. Supports 1 placeholder).
        """

        type: ComponentType = dataclasses.field(
            default=ComponentType.HEADER, init=False, repr=False
        )
        format: HeaderFormatType = dataclasses.field(
            default=HeaderFormatType.TEXT, init=False, repr=False
        )
        text: str

        def to_dict(self, placeholder: tuple[str, str] = None) -> dict[str, str | None]:
            formatted_text, examples = _get_examples_from_placeholders(
                self.text, *(placeholder if placeholder else ())
            )
            return dict(
                type=self.type.value,
                format=self.format.value,
                text=formatted_text,
                **(dict(example=dict(header_text=examples)) if examples else {}),
            )

    @dataclasses.dataclass(slots=True)
    class Image(NewTemplateHeaderABC):
        """
        Represents an image header.

        Example:
            >>> from pywa.types.template import NewTemplate
            >>> NewTemplate.Image(example="2:c2FtcGxl...")

        Attributes:
            example: An image handles (Use the `Resumable Upload API
             <https://developers.facebook.com/docs/graph-api/guides/upload>`_ to upload the image)
        """

        type: ComponentType = dataclasses.field(
            default=ComponentType.HEADER, init=False, repr=False
        )
        format: HeaderFormatType = dataclasses.field(
            default=HeaderFormatType.IMAGE, init=False, repr=False
        )
        example: str

        def to_dict(self) -> dict[str, Any]:
            return dict(
                type=self.type.value,
                format=self.format.value,
                example=dict(header_handle=(self.example,)),
            )

    @dataclasses.dataclass(slots=True)
    class Video(NewTemplateHeaderABC):
        """
        Represents a video header.

        Example:
            >>> from pywa.types import NewTemplate
            >>> NewTemplate.Video(example="2:c2FtcGxl...")

        Attributes:
            example: A video handle (Use the `Resumable Upload API
             <https://developers.facebook.com/docs/graph-api/guides/upload>`_ to upload the video)
        """

        type: ComponentType = dataclasses.field(
            default=ComponentType.HEADER, init=False, repr=False
        )
        format: HeaderFormatType = dataclasses.field(
            default=HeaderFormatType.VIDEO, init=False, repr=False
        )
        example: str

        def to_dict(self) -> dict[str, Any]:
            return dict(
                type=self.type.value,
                format=self.format.value,
                example=dict(header_handle=(self.example,)),
            )

    @dataclasses.dataclass(slots=True)
    class Document(NewTemplateHeaderABC):
        """
        Represents a document header.

        Example:
            >>> from pywa.types import NewTemplate
            >>> NewTemplate.Document(example="2:c2FtcGxl...")

        Attributes:
            example: A document handle (Use the `Resumable Upload API
             <https://developers.facebook.com/docs/graph-api/guides/upload>`_ to upload the document)
        """

        type: ComponentType = dataclasses.field(
            default=ComponentType.HEADER, init=False, repr=False
        )
        format: HeaderFormatType = dataclasses.field(
            default=HeaderFormatType.DOCUMENT, init=False, repr=False
        )
        example: str

        def to_dict(self) -> dict[str, Any]:
            return dict(
                type=self.type.value,
                format=self.format.value,
                example=dict(header_handle=(self.example,)),
            )

    @dataclasses.dataclass(slots=True)
    class Location(NewTemplateHeaderABC):
        """
        Location headers appear as generic maps at the top of the template and are useful for order tracking, delivery
        updates, ride hailing pickup/dropoff, locating physical stores, etc. When tapped, the app user's default map app
        will open and load the specified location. Locations are specified when you send the template.

        - Location headers can only be used in templates categorized as ``UTILITY`` or ``MARKETING``. Real-time locations are not supported.

        Example:
            >>> from pywa.types import NewTemplate
            >>> NewTemplate.Location()
        """

        type: ComponentType = dataclasses.field(
            default=ComponentType.HEADER, init=False, repr=False
        )
        format: HeaderFormatType = dataclasses.field(
            default=HeaderFormatType.LOCATION, init=False, repr=False
        )

        def to_dict(self) -> dict[str, str]:
            return dict(
                type=self.type.value,
                format=self.format.value,
            )

    @dataclasses.dataclass(slots=True)
    class Body(NewTemplateComponentABC):
        """
        Represents a template body.

        Example:

            >>> from pywa.types import NewTemplate
            >>> NewTemplate.Body(
            ...     text='Shop now through {the end of August} and '
            ...         'use code {25OF} to get {25%} off of all merchandise',
            ... )

        Attributes:
            text: Text to send with the body (Up to 1024 characters. Supports multiple placeholders).
        """

        type: ComponentType = dataclasses.field(
            default=ComponentType.BODY, init=False, repr=False
        )
        text: str

        def to_dict(self, placeholder: tuple[str, str] = None) -> dict[str, str | None]:
            formatted_text, examples = _get_examples_from_placeholders(
                self.text, *(placeholder if placeholder else ())
            )
            return dict(
                type=self.type.value,
                text=formatted_text,
                **(dict(example=dict(body_text=(examples,))) if examples else {}),
            )

    @dataclasses.dataclass(slots=True)
    class AuthBody:
        """
        Represents the configuration for an authentication template.

        Example:

            >>> from pywa.types import NewTemplate
            >>> NewTemplate.AuthBody(
            ...     code_expiration_minutes=5,
            ...     add_security_recommendation=True
            ... )

        Attributes:
            code_expiration_minutes: Indicates number of minutes the password or code is valid. If omitted, the code
             expiration warning will not be displayed in the delivered message. (Minimum ``1``, maximum ``90``).
            add_security_recommendation: Set to ``True`` if you want the template to include the string, ``"For your security,
             do not share this code"``. Set to ``False`` to exclude the string. Defaults to ``False``.
        """

        code_expiration_minutes: int | None = None
        add_security_recommendation: bool = False

    @dataclasses.dataclass(slots=True)
    class Footer(NewTemplateComponentABC):
        """
        Represents a template footer.

        Example:
            >>> from pywa.types import NewTemplate
            >>> NewTemplate.Footer(text='⚡ Powered by PyWa')

        Attributes:
            text: Text to send with the footer (Up to 60 characters, no placeholders allowed).
        """

        type: ComponentType = dataclasses.field(
            default=ComponentType.FOOTER, init=False, repr=False
        )
        text: str

        def to_dict(self) -> dict[str, str | None]:
            return dict(type=self.type.value, text=self.text)

    @dataclasses.dataclass(slots=True)
    class PhoneNumberButton(NewButtonABC):
        """
        Phone number buttons call the specified business phone number when tapped by the app user.
        Templates are limited to one phone number button.

        - It seems that no need to provide the phone button when sending the template, it will be added automatically

        Example:
            >>> from pywa.types import NewTemplate
            >>> NewTemplate.PhoneNumberButton(title='Call Us', phone_number='12125552368')

        Attributes:
            title: Button text (Up to 25 characters, no placeholders allowed).
            phone_number: Alphanumeric string. Business phone number to be (display phone number)
             called when the user taps the button (Up to 20 characters, no placeholders allowed).
        """

        type: ButtonType = dataclasses.field(
            default=ButtonType.PHONE_NUMBER, init=False, repr=False
        )
        title: str
        phone_number: int | str

        def to_dict(self, placeholder: None = None) -> dict[str, str]:
            return dict(
                type=self.type.value,
                text=self.title,
                phone_number=str(self.phone_number),
            )

    @dataclasses.dataclass(slots=True)
    class UrlButton(NewButtonABC):
        """
        URL buttons load the specified URL in the device's default web browser when tapped by the app user.
        Templates are limited to two URL buttons.

        Example:
            >>> from pywa.types import NewTemplate
            >>> NewTemplate.UrlButton(
            ...     title='Log In',
            ...     url='https://x.com/login?email={john@example}'
            ... )

        Attributes:
            title: Button text (Up to 25 characters, supports 1 variable).
            url: URL to be loaded when the user taps the button (Up to 2000 characters, supports 1 placeholder, which
             be appended to the end of the URL).
        """

        type: ComponentType = dataclasses.field(
            default=ButtonType.URL, init=False, repr=False
        )
        title: str
        url: str

        def to_dict(self, placeholder: tuple[str, str] = None) -> dict[str, str]:
            (
                formatted_title,
                title_examples,
            ) = _get_examples_from_placeholders(
                self.title, *(placeholder if placeholder else ())
            )
            (
                formatted_url,
                url_examples,
            ) = _get_examples_from_placeholders(
                self.url, *(placeholder if placeholder else ())
            )
            examples = title_examples + url_examples
            return dict(
                type=self.type.value,
                text=formatted_title,
                url=formatted_url,
                **(dict(example=examples if examples else {})),
            )

    @dataclasses.dataclass(slots=True)
    class QuickReplyButton(NewButtonABC):
        """
        Quick reply buttons are custom text-only buttons that immediately message you with the specified text string when
        tapped by the app user. A common use case-case is a button that allows your customer to easily opt-out of any
        marketing messages.

            - You will provide the callback data when you send the template.

        Templates are limited to 10 quick reply buttons. If using quick reply buttons with other buttons, buttons must be
        organized into two groups: quick reply buttons and non-quick reply buttons. If grouped incorrectly, the API will
        return an error indicating an invalid combination.

        Example:

            >>> from pywa.types import NewTemplate
            >>> NewTemplate.QuickReplyButton(text='Unsubscribe from Promos')
            >>> NewTemplate.QuickReplyButton(text='Unsubscribe from All')

        Attributes:
            text: The text to send when the user taps the button (Up to 25 characters, no placeholders allowed).
        """

        type: ComponentType = dataclasses.field(
            default=ButtonType.QUICK_REPLY, init=False, repr=False
        )
        text: str

        def to_dict(self, placeholder: None = None) -> dict[str, str]:
            return dict(type=self.type.value, text=self.text)

    @dataclasses.dataclass(slots=True)
    class OTPButton(NewButtonABC):
        """
        Represents a button that can be used to send an OTP.

        Example for ONE_TAP:

            >>> from pywa.types import NewTemplate
            >>> NewTemplate.OTPButton(
            ...     otp_type=NewTemplate.OTPButton.OtpType.ONE_TAP,
            ...     title='Copy Code',
            ...     autofill_text='Autofill',
            ...     package_name='com.example.app',
            ...     signature_hash='1234567890ABCDEF1234567890ABCDEF12345678'
            ... )

        Example for COPY_CODE:

            >>> from pywa.types import NewTemplate
            >>> NewTemplate.OTPButton(
            ...     otp_type=NewTemplate.OTPButton.OtpType.COPY_CODE,
            ...     title='Copy Code'
            ... )


        Attributes:
            otp_type: Type of OTP to send. A copy code button copies the one-time password or code to the user's clipboard.
             The user can then manually switch to your app and paste the password or code into your app's interface. A
             one-tap autofill button automatically loads and passes your app the one-time password or code.
            title: Copy code button text. If omitted, the text will default to a pre-set value localized to the template's
             language. For example, ``Copy Code`` for English (US).
            autofill_text: One-tap autofill button text. If omitted, the text will default to a pre-set value localized to
             the template's language. For example, ``Autofill`` for English (US).
            package_name: Android package name of your app. Required if ``otp_type`` is ``OtpType.ONE_TAP`` or ``OtpType.ZERO_TAP``.
            signature_hash: Your app signing key hash. See `App Signing Key Hash
             <https://developers.facebook.com/docs/whatsapp/business-management-api/authentication-templates#app-signing-key-hash>`_
             for more information. Required if ``otp_type`` is ``OtpType.ONE_TAP`` or ``OtpType.ZERO_TAP``.
            zero_tap_terms_accepted: Acceptance of the WhatsApp Business API Terms of Service. Required if ``otp_type`` is
             ``OtpType.ZERO_TAP``. Defaults to ``True``.
        """

        type: ComponentType = dataclasses.field(
            default=ButtonType.OTP, init=False, repr=False
        )
        otp_type: OtpType
        title: str | None = None
        autofill_text: str | None = None
        package_name: str | None = None
        signature_hash: str | None = None
        zero_tap_terms_accepted: bool = True

        class OtpType(utils.StrEnum):
            """
            The type of the button

            Attributes:
                COPY_CODE: Copy code button copies the one-time password or code to the user's clipboard.
                ONE_TAP: One-tap autofill button automatically loads and passes your app the one-time password or code.
                ZERO_TAP: Zero-tap autofill button automatically loads and passes your app the one-time password or code
            """

            COPY_CODE = "COPY_CODE"
            ONE_TAP = "ONE_TAP"
            ZERO_TAP = "ZERO_TAP"

        def __post_init__(self):
            if self.otp_type in (self.OtpType.ONE_TAP, self.OtpType.ZERO_TAP) and not (
                self.package_name and self.signature_hash
            ):
                raise ValueError(
                    "`package_name` and `signature_hash` are required for ONE_TAP and ZERO_TAP"
                )

        def to_dict(self, placeholder: None = None) -> dict[str, str | None]:
            base = dict(
                type=self.type.value,
                otp_type=self.otp_type.value,
                text=self.title,
            )
            if self.otp_type in (self.OtpType.ONE_TAP, self.OtpType.ZERO_TAP):
                base.update(
                    package_name=self.package_name,
                    signature_hash=self.signature_hash,
                    **(
                        dict(autofill_text=self.autofill_text)
                        if self.autofill_text
                        else {}
                    ),
                    **(
                        dict(zero_tap_terms_accepted=self.zero_tap_terms_accepted)
                        if self.otp_type == self.OtpType.ZERO_TAP
                        else {}
                    ),
                )
            return base

    @dataclasses.dataclass(slots=True)
    class MPMButton(NewButtonABC):
        """
        Represents a button that can be used to send multi-product message (MPM)
            - This button required providing header to the new template!

        Example:
            >>> from pywa.types import NewTemplate
            >>> NewTemplate.MPMButton()
        """

        type: ComponentType = dataclasses.field(
            default=ButtonType.MPM, init=False, repr=False
        )

        def to_dict(self, placeholder: None = None) -> dict[str, str]:
            return dict(
                type=self.type.value, text="View items"
            )  # required text for MPM button

    @dataclasses.dataclass(slots=True)
    class CatalogButton(NewButtonABC):
        """
        Represent a button that can be used to display a catalog.

        Example:
            >>> from pywa.types import NewTemplate
            >>> NewTemplate.CatalogButton()

        """

        type: ComponentType = dataclasses.field(
            default=ButtonType.CATALOG, init=False, repr=False
        )

        def to_dict(self, placeholder: None = None) -> dict[str, str]:
            return dict(
                type=self.type.value, text="View catalog"
            )  # required text for catalog button

    @dataclasses.dataclass(slots=True)
    class CopyCodeButton(NewButtonABC):
        """
        Represents a button that can be used to copy coupon codes to the user's clipboard.
            - Only one copy code button is allowed per template.
            - Do not use this button to send one-time passwords (OTPs), Use :class:`NewTemplate.OTPButton` instead.

        Example:
            >>> from pywa.types import NewTemplate
            >>> NewTemplate.CopyCodeButton(example='250FF')


        Attributes:
            example: An example of the coupon code (Up to 15 characters).
        """

        type: ComponentType = dataclasses.field(
            default=ButtonType.COPY_CODE, init=False, repr=False
        )
        example: str

        def to_dict(self, placeholder: tuple[str, str] = None) -> dict[str, str | None]:
            return dict(type=self.type.value, example=self.example)

    @dataclasses.dataclass(slots=True)
    class FlowButton(NewButtonABC):
        """
        Represents a button that can be used to send a message with a flow.

        Example:
            >>> from pywa.types import NewTemplate, FlowActionType
            >>> NewTemplate.FlowButton(
            ...     flow_id='123456789',
            ...     title='Feedback',
            ...     flow_action=FlowActionType.NAVIGATE,
            ...     navigate_screen='SURVEY_SCREEN'
            ... )

        Attributes:
            flow_id: The flow ID to send.
            title: The button title.
            flow_action: The flow action (``NAVIGATE`` or ``DATA_EXCHANGE``).
            navigate_screen: The screen to navigate to (required if ``flow_action`` is ``NAVIGATE``).
        """

        type: ButtonType = dataclasses.field(
            default=ButtonType.QUICK_REPLY, init=False, repr=False
        )
        title: str
        flow_id: str | int
        flow_action: Literal[FlowActionType.NAVIGATE, FlowActionType.DATA_EXCHANGE]
        navigate_screen: str | None = None

        def __post_init__(self):
            if self.flow_action == FlowActionType.NAVIGATE and not self.navigate_screen:
                raise ValueError("`navigate_screen` is required for FLOW with NAVIGATE")

        def to_dict(self, placeholder: None = None) -> dict[str, str]:
            return dict(
                type=self.type.value,
                text=self.title,
                flow_id=str(self.flow_id),
                flow_action=self.flow_action,
                **(
                    {"navigate_screen": self.navigate_screen}
                    if self.navigate_screen
                    else {}
                ),
            )


class ParamType(utils.StrEnum):
    TEXT = "text"
    CURRENCY = "currency"
    DATE_TIME = "date_time"
    DOCUMENT = "document"
    IMAGE = "image"
    VIDEO = "video"
    LOCATION = "location"
    BUTTON = "button"


class ComponentABC(abc.ABC):
    @property
    @abc.abstractmethod
    def type(self) -> ParamType: ...


@dataclasses.dataclass(slots=True)
class Template:
    """
    The template to send.

    The template message is basically a message that contains the values of the template placeholders.

    If in the template created with body that contains 3 variables, then the message must contain 3 values for the
    placeholders.

    >>> from pywa.types import Template as Temp
    >>> Temp(
    ...     name='buy_new_iphone_x',
    ...     language=Temp.Language.ENGLISH_US,
    ...     header=TextValue(value='15'),
    ...     body=[
    ...         Temp.TextValue(value='John Doe'),
    ...         Temp.TextValue(value='WA_IPHONE_15'),
    ...         Temp.TextValue(value='15%'),
    ...     ],
    ...     buttons=[
    ...         Temp.UrlButtonValue(value='iphone15'),
    ...         Temp.QuickReplyButtonData(data='unsubscribe_from_marketing_messages'),
    ...         Temp.QuickReplyButtonData(data='unsubscribe_from_all_messages'),
    ...     ],
    ... )

    Example for Authentication template:

    >>> from pywa.types import Template as Temp
    >>> Temp(
    ...     name='auth_with_otp',
    ...     language=Temp.Language.ENGLISH_US,
    ...     buttons=Temp.OTPButtonCode(code='123456'),
    ... )


    Attributes:
        name: The name of the template.
        language: The language of the template (See `Template language and locale code
            <https://developers.facebook.com/docs/whatsapp/business-management-api/message-templates/supported-languages>`_).
        body: The body of the template, which can include one or more text variables.
        header: The header of the template (required base on the template you are sending).
        buttons: The buttons to send with the template (required base on the template you are sending).
    """

    name: str
    language: Language | str
    body: Iterable[TextValue | Currency | DateTime] | None = None
    header: TextValue | Document | Image | Video | Location | None = None
    buttons: (
        Iterable[QuickReplyButtonData | UrlButtonValue]
        | OTPButtonCode
        | MPMButton
        | CatalogButton
        | CopyCodeButton
        | FlowButton
        | None
    ) = None

    def __post_init__(self):
        if isinstance(self.buttons, self.OTPButtonCode):
            self.body = (
                self.TextValue(value=self.buttons.code),
            )  # auth template required the code also in the body

    def to_dict(self, is_header_url: bool = False) -> dict[str, Any]:
        return dict(
            name=self.name,
            language=dict(code=str(self.language)),
            components=tuple(
                comp
                for comp in (
                    dict(
                        type=ComponentType.BODY.value,
                        parameters=tuple(
                            component.to_dict() for component in self.body
                        ),
                    )
                    if self.body
                    else None,
                    dict(
                        type=ComponentType.HEADER.value,
                        parameters=(self.header.to_dict(is_header_url),),
                    )
                    if self.header
                    else None,
                    *(
                        (
                            dict(
                                type=b.type.value,
                                sub_type=b.sub_type.value,
                                index=idx,
                                parameters=(b.to_dict(),),
                            )
                            for idx, b in enumerate(
                                self.buttons
                                if isinstance(self.buttons, Iterable)
                                else (self.buttons,)  # case of OTPButtonCode
                            )
                        )
                        if self.buttons is not None
                        else ()
                    ),
                )
                if comp is not None
            ),
        )

    Language = Language

    @dataclasses.dataclass(slots=True)
    class TextValue(ComponentABC):
        """
        Represents a value to assign to a placeholder in a template.

        Example:
            >>> from pywa.types import Template
            >>> Template.TextValue(var='John Doe')  # The template was created with 'Hello, {John}!'

        Attributes:
            value: The value to assign to the placeholder.
        """

        type: ParamType = dataclasses.field(
            default=ParamType.TEXT, init=False, repr=False
        )
        value: str

        def to_dict(self, is_url: None = None) -> dict[str, str]:
            return dict(type=self.type.value, text=self.value)

    @dataclasses.dataclass(slots=True)
    class Currency(ComponentABC):
        """
        Represents a currency variable.

        Attributes:
            fallback_value: Default text if localization fails.
            code: ISO 4217 currency code (e.g. USD, EUR, etc.).
            amount_1000: Amount multiplied by 1000.
        """

        type: ParamType = dataclasses.field(
            default=ParamType.CURRENCY, init=False, repr=False
        )
        fallback_value: str
        code: str
        amount_1000: int

        def to_dict(self) -> dict[str, str]:
            return dict(
                type=self.type.value,
                currency=dict(
                    fallback_value=self.fallback_value,
                    code=self.code,
                    amount_1000=self.amount_1000,
                ),
            )

    @dataclasses.dataclass(slots=True)
    class DateTime(ComponentABC):
        """
        Represents a date time variable.

        Attributes:
            fallback_value: Default text if localization fails.
        """

        type: ParamType = dataclasses.field(
            default=ParamType.DATE_TIME, init=False, repr=False
        )
        fallback_value: str

        def to_dict(self) -> dict[str, str]:
            return dict(
                type=self.type.value,
                date_time=dict(
                    fallback_value=self.fallback_value,
                ),
            )

    @dataclasses.dataclass(slots=True)
    class Document(ComponentABC):
        """
        Represents a document.

        Attributes:
            document: The document to send (PDF only. either a media ID, URL, file path, bytes, or an open file object).
            caption: The caption of the document.
            filename: The filename of the document (Required if sending bytes or an open file object).
        """

        type: ParamType = dataclasses.field(
            default=ParamType.DOCUMENT, init=False, repr=False
        )
        document: str | pathlib.Path | bytes | BinaryIO
        caption: str | None = None
        filename: str | None = None

        def to_dict(self, is_url: bool) -> dict[str, str]:
            return dict(
                type=self.type.value,
                document={"link" if is_url else "id": self.document}
                | (dict(caption=self.caption) if self.caption else {})
                | (dict(filename=self.filename) if self.filename else {}),
            )

    @dataclasses.dataclass(slots=True)
    class Image(ComponentABC):
        """
        Represents an image.

        Attributes:
            image: The image to send (either a media ID, URL, file path, bytes, or an open file object).
            caption: The caption of the image.
            mime_type: The mime type of the image (Required if sending bytes or an open file object).
        """

        type: ParamType = dataclasses.field(
            default=ParamType.IMAGE, init=False, repr=False
        )
        image: str | pathlib.Path | bytes | BinaryIO
        caption: str | None = None
        mime_type: str | None = None

        def to_dict(self, is_url: bool) -> dict[str, str]:
            return dict(
                type=self.type.value,
                image={"link" if is_url else "id": self.image}
                | (dict(caption=self.caption) if self.caption else {}),
            )

    @dataclasses.dataclass(slots=True)
    class Video(ComponentABC):
        """
        Represents a video.

        Attributes:
            video: The video to send (either a media ID, URL, file path, bytes, or an open file object).
            caption: The caption of the video.
            mime_type: The mime type of the video (Required if sending bytes or an open file object).
        """

        type: ParamType = dataclasses.field(
            default=ParamType.VIDEO, init=False, repr=False
        )
        video: str | pathlib.Path | bytes | BinaryIO
        caption: str | None = None
        mime_type: str | None = None

        def to_dict(self, is_url: bool) -> dict[str, str]:
            return dict(
                type=self.type.value,
                video={"link" if is_url else "id": self.video}
                | (dict(caption=self.caption) if self.caption else {}),
            )

    @dataclasses.dataclass(slots=True)
    class Location(ComponentABC):
        """
        Represents a location.

        Attributes:
            latitude: The latitude of the location.
            longitude: The longitude of the location.
            name: The name of the location.
            address: The address of the location.
        """

        type: ParamType = dataclasses.field(
            default=ParamType.LOCATION, init=False, repr=False
        )
        latitude: float
        longitude: float
        name: str | None = None
        address: str | None = None

        def to_dict(self) -> dict[str, str]:
            return dict(
                type=self.type.value,
                location=dict(
                    latitude=self.latitude,
                    longitude=self.longitude,
                    **(dict(name=self.name) if self.name else {}),
                    **(dict(address=self.address) if self.address else {}),
                ),
            )

    @dataclasses.dataclass(slots=True)
    class QuickReplyButtonData(ComponentABC):
        """
        Represents a quick reply button.

        Attributes:
            data: The data to send when the user taps the button
             (you can listen for this data with @on_callback_button decorator).
        """

        type: ParamType = dataclasses.field(
            default=ParamType.BUTTON, init=False, repr=False
        )
        sub_type: ButtonType = dataclasses.field(
            default=ButtonType.QUICK_REPLY, init=False, repr=False
        )
        data: CallbackDataT

        def to_dict(self) -> dict[str, str]:
            return dict(
                type="payload",
                payload=_resolve_callback_data(self.data),
            )

    @dataclasses.dataclass(slots=True)
    class UrlButtonValue(ComponentABC):
        """
        Represents a URL button variable.

        Example:
            >>> from pywa.types import Template
            >>> Template.UrlButtonValue(value='COUPON123')  # The template was created with 'https://example.com/shop/{COUPON12}'

        Attributes:
            value: The value to assign to the variable in the template (appended to the end of the URL).
        """

        type: ParamType = dataclasses.field(
            default=ParamType.BUTTON, init=False, repr=False
        )
        sub_type: ButtonType = dataclasses.field(
            default=ButtonType.URL, init=False, repr=False
        )
        value: str

        def to_dict(self) -> dict[str, str]:
            return dict(type="text", text=self.value)

    @dataclasses.dataclass(slots=True)
    class OTPButtonCode(ComponentABC):
        """
        Represents an OTP button variable.

        Example:
            >>> from pywa.types import Template
            >>> Template.OTPButtonCode(value='123456')

        Attributes:
            code: The code to copy or autofill when the user taps the button.
        """

        type: ParamType = dataclasses.field(
            default=ParamType.BUTTON, init=False, repr=False
        )
        sub_type: ButtonType = dataclasses.field(
            default=ButtonType.URL, init=False, repr=False
        )
        code: str

        def to_dict(self) -> dict[str, str]:
            return dict(type="text", text=self.code)

    @dataclasses.dataclass(slots=True)
    class MPMButton(ComponentABC):
        """
        Represent a multi-product message (MPM) button

        Example:
            >>> from pywa.types import Template
            >>> Template.MPMButton(
            ...     thumbnail_product_sku='IPHONE_15',
            ...     product_sections=[
            ...         ProductsSection(
            ...             title='Smartphones',
            ...             skus=['IPHONE_15', 'GALAXY_S23'],
            ...         ),
            ...         ProductsSection(
            ...             title='Laptops',
            ...             skus=['MACBOOKPRO', 'SURFACEPRO'],
            ...         ),
            ...     ],
            ... )

        Attributes:
            thumbnail_product_sku: The thumbnail of this item will be used as the template message's header image.
            product_sections: The product sections to send with the template.
        """

        type: ParamType = dataclasses.field(
            default=ParamType.BUTTON, init=False, repr=False
        )
        sub_type: ButtonType = dataclasses.field(
            default=ButtonType.MPM, init=False, repr=False
        )
        thumbnail_product_sku: str
        product_sections: Iterable[ProductsSection]

        def to_dict(self) -> dict[str, str]:
            return dict(
                type="action",
                action=dict(
                    thumbnail_product_retailer_id=self.thumbnail_product_sku,
                    sections=tuple(
                        section.to_dict() for section in self.product_sections
                    ),
                ),
            )

    @dataclasses.dataclass(slots=True)
    class CatalogButton(ComponentABC):
        """
        Represent a catalog button

        Example:
            >>> from pywa.types import Template
            >>> Template.CatalogButton(thumbnail_product_sku='IPHONE_15')

        Attributes:
            thumbnail_product_sku: The thumbnail of this item will be used as the message's header image. if not
                provided, the product image of the first item in your catalog will be used.
        """

        type: ParamType = dataclasses.field(
            default=ParamType.BUTTON, init=False, repr=False
        )
        sub_type: ButtonType = dataclasses.field(
            default=ButtonType.CATALOG, init=False, repr=False
        )
        thumbnail_product_sku: str | None = None

        def to_dict(self) -> dict[str, str]:
            return dict(
                type="action",
                action=dict(
                    thumbnail_product_retailer_id=self.thumbnail_product_sku,
                ),
            )

    @dataclasses.dataclass(slots=True)
    class CopyCodeButton(ComponentABC):
        """
        Represents a copy code button variable (copies the code to the user's clipboard).

        Example:
            >>> from pywa.types import Template
            >>> Template.CopyCodeButton(code='COUPON123')

        Attributes:
            code: The code to copy when the user taps the button.
        """

        type: ParamType = dataclasses.field(
            default=ParamType.BUTTON, init=False, repr=False
        )
        sub_type: ButtonType = dataclasses.field(
            default=ButtonType.COPY_CODE, init=False, repr=False
        )
        code: str

        def to_dict(self) -> dict[str, str]:
            return dict(type="coupon_code", coupon_code=self.code)

    @dataclasses.dataclass(slots=True)
    class FlowButton(ComponentABC):
        """
        Represents a flow button.

        Example:
            >>> from pywa.types import Template, FlowActionType
            >>> Template.FlowButton(
            ...     flow_token="AAAAAAAAAA",
            ...     flow_action_data={"is_name_required": True}
            ... )

        Attributes:
            flow_token: The flow token to send to identify the flow when exchanging data. Optional, default is "unused"
            flow_action_data: The data to pass to the screen that specifies when creating the template.
        """

        type: ParamType = dataclasses.field(
            default=ParamType.BUTTON, init=False, repr=False
        )
        sub_type: ButtonType = dataclasses.field(
            default=ButtonType.FLOW, init=False, repr=False
        )
        flow_token: str | None = None
        flow_action_data: dict[str, Any] | None = None

        def to_dict(self) -> dict[str, str]:
            return dict(
                type="action",
                action=dict(
                    **({"flow_token": self.flow_token} if self.flow_token else {}),
                    **(
                        {"flow_action_data": self.flow_action_data}
                        if self.flow_action_data
                        else {}
                    ),
                ),
            )


@dataclasses.dataclass(slots=True, frozen=True, kw_only=True)
class TemplateStatus(BaseUpdate):
    """
    Represents status of a template.

    Attributes:
        id: ID of Whatsapp Business Accounts this update belongs to.
        timestamp: Timestamp of the update.
        event: The event that occurred (the template was approved, rejected, etc.).
        message_template_id: The ID of the template.
        message_template_name: The name of the template.
        message_template_language: The language of the template.
        reason: The reason the template was rejected (if applicable).
        disable_date: The date the template was disabled (if applicable).
        other_info: Additional information about the template (if applicable).
    """

    id: str
    timestamp: datetime.datetime
    event: TemplateEvent
    message_template_id: int
    message_template_name: str
    message_template_language: str
    reason: TemplateRejectionReason
    disable_date: str | None = None
    other_info: str | None = None

    @classmethod
    def from_update(cls, client: WhatsApp, update: dict) -> TemplateStatus:
        value = (data := update["entry"][0])["changes"][0]["value"]
        return cls(
            _client=client,
            raw=update,
            id=data["id"],
            timestamp=datetime.datetime.fromtimestamp(data["time"]),
            event=cls.TemplateEvent(value["event"]),
            message_template_id=value["message_template_id"],
            message_template_name=value["message_template_name"],
            message_template_language=value["message_template_language"],
            reason=cls.TemplateRejectionReason(
                str(value.get("reason"))
            ),  # _missing_(str(None)) -> .NONE
            disable_date=value.get("disable_date"),
            other_info=(
                str((oi := value["other_info"]).get("title"))
                + ": "
                + str(oi.get("description"))
            )
            if "other_info" in value
            else None,
        )

    class TemplateEvent(utils.StrEnum):
        """
        The event that occurred (the template was approved, rejected, etc.).

        Attributes:
            APPROVED: The template was approved.
            DISABLED: The template was disabled.
            IN_APPEAL: The template is in appeal.
            PENDING: The template is pending.
            REINSTATED: The template was reinstated.
            REJECTED: The template was rejected.
            PENDING_DELETION: The template is pending deletion.
            FLAGGED: The template was flagged.
            PAUSED: The template was paused.
            UNKNOWN: Unknown event.
        """

        APPROVED = "APPROVED"
        DISABLED = "DISABLED"
        IN_APPEAL = "IN_APPEAL"
        PENDING = "PENDING"
        REINSTATED = "REINSTATED"
        REJECTED = "REJECTED"
        PENDING_DELETION = "PENDING_DELETION"
        FLAGGED = "FLAGGED"
        PAUSED = "PAUSED"
        UNKNOWN = "UNKNOWN"

        @classmethod
        def _missing_(cls, value: str) -> TemplateStatus.TemplateEvent:
            _logger.warning("Unknown template event: %s. Defaulting to UNKNOWN", value)
            return cls.UNKNOWN

    class TemplateRejectionReason(utils.StrEnum):
        """
        The reason the template was rejected (if applicable).

        Attributes:
            ABUSIVE_CONTENT: The template was rejected because it contained abusive content.
            INCORRECT_CATEGORY: The template was rejected because it was in the wrong category.
            INVALID_FORMAT: The template was rejected because it was in the wrong format.
            SCAM: The template was rejected because it was a scam.
            NONE: The template was not rejected.
        """

        ABUSIVE_CONTENT = "ABUSIVE_CONTENT"
        INCORRECT_CATEGORY = "INCORRECT_CATEGORY"
        INVALID_FORMAT = "INVALID_FORMAT"
        SCAM = "SCAM"
        NONE = "NONE"

        @classmethod
        def _missing_(cls, value: str):
            _logger.warning(
                "Unknown template rejection reason: %s. Defaulting to NONE", value
            )
            return cls.NONE
